#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Final, Iterable, Union

import torch

from torchoutil.core.semver import Version
from torchoutil.pyoutil.functools import identity
from torchoutil.pyoutil.importlib import package_is_available

_EXTRAS_PACKAGES = (
    "colorlog",
    "h5py",
    "numpy",
    "omegaconf",
    "pandas",
    "safetensors",
    "scipy",
    "tensorboard",
    "tqdm",
    "yaml",
)
_EXTRA_AVAILABLE = {name: package_is_available(name) for name in _EXTRAS_PACKAGES}


_COLORLOG_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["colorlog"]
_H5PY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["h5py"]
_NUMPY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["numpy"]
_OMEGACONF_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["omegaconf"]
_PANDAS_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["pandas"]
_SAFETENSORS_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["safetensors"]
_SCIPY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["scipy"]
_TENSORBOARD_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["tensorboard"]
_TQDM_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["tqdm"]
_YAML_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["yaml"]


def requires_packages(packages: Union[str, Iterable[str]]) -> Callable:
    if isinstance(packages, str):
        packages = [packages]
    else:
        packages = list(packages)

    missing = [pkg for pkg in packages if not package_is_available(pkg)]
    if len(missing) == 0:
        return identity

    prefix = "\n - "
    missing_str = prefix.join(missing)
    msg = (
        f"Cannot use/import objects because the following optionals dependencies are missing:"
        f"{prefix}{missing_str}\n"
        f"Please install them using `pip install torchoutil[extras]`."
    )
    raise ImportError(msg)


def torch_version_ge_1_13() -> bool:
    version_str = str(torch.__version__)
    version = Version.from_str(version_str)
    return version >= Version("1.13.0")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch.types import *

from torchoutil.core.dtype_enum import DTypeEnum
from torchoutil.core.get import DeviceLike, DTypeLike, GeneratorLike
from torchoutil.pyoutil.typing import *

from .guards import (
    is_bool_tensor,
    is_bool_tensor1d,
    is_builtin_number,
    is_builtin_scalar,
    is_integral_dtype,
    is_integral_tensor,
    is_integral_tensor1d,
    is_iterable_tensor,
    is_list_tensor,
    is_number_like,
    is_numpy_number_like,
    is_numpy_scalar_like,
    is_scalar_like,
    is_tensor0d,
    is_tensor_like,
    is_tuple_tensor,
)
from .tensor_subclasses import (
    BoolTensor,
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    ByteTensor,
    ByteTensor0D,
    ByteTensor1D,
    ByteTensor2D,
    ByteTensor3D,
    CDoubleTensor,
    CDoubleTensor0D,
    CDoubleTensor1D,
    CDoubleTensor2D,
    CDoubleTensor3D,
    CFloatTensor,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    CHalfTensor,
    CHalfTensor0D,
    CHalfTensor1D,
    CHalfTensor2D,
    CHalfTensor3D,
    CharTensor,
    CharTensor0D,
    CharTensor1D,
    CharTensor2D,
    CharTensor3D,
    ComplexFloatingTensor,
    ComplexFloatingTensor0D,
    ComplexFloatingTensor1D,
    ComplexFloatingTensor2D,
    ComplexFloatingTensor3D,
    DoubleTensor,
    DoubleTensor0D,
    DoubleTensor1D,
    DoubleTensor2D,
    DoubleTensor3D,
    FloatingTensor,
    FloatingTensor0D,
    FloatingTensor1D,
    FloatingTensor2D,
    FloatingTensor3D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    HalfTensor,
    HalfTensor0D,
    HalfTensor1D,
    HalfTensor2D,
    HalfTensor3D,
    IntTensor,
    IntTensor0D,
    IntTensor1D,
    IntTensor2D,
    IntTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    ShortTensor,
    ShortTensor0D,
    ShortTensor1D,
    ShortTensor2D,
    ShortTensor3D,
    SignedIntegerTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
    SignedIntegerTensor2D,
    SignedIntegerTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Iterable, Tuple, Union, overload

import torch
from torch import Tensor
from typing_extensions import TypeIs

from torchoutil.core.get import DeviceLike, DTypeLike, get_device, get_dtype
from torchoutil.core.packaging import torch_version_ge_1_13
from torchoutil.extras.numpy.definitions import NumpyNumberLike, NumpyScalarLike, np
from torchoutil.pyoutil import (
    BuiltinScalar,
    get_current_fn_name,
    is_builtin_scalar,
    prod,
)


def to_numpy(
    x: Union[Tensor, np.ndarray, Iterable, BuiltinScalar],
    *,
    dtype: Union[str, np.dtype, None] = None,
    force: bool = False,
) -> np.ndarray:
    """Convert input to numpy array. Works with any arbitrary object."""
    if isinstance(x, Tensor):
        return tensor_to_numpy(x, dtype=dtype, force=force)
    else:
        return np.asarray(x, dtype=dtype)  # type: ignore


def tensor_to_numpy(
    x: Tensor,
    *,
    dtype: Union[str, np.dtype, None] = None,
    force: bool = False,
) -> np.ndarray:
    """Convert PyTorch tensor to numpy array."""
    if torch_version_ge_1_13():
        kwargs = dict(force=force)
    elif not force:
        kwargs = dict()
    else:
        msg = f"Invalid argument {force=} for {get_current_fn_name()}. (expected True because torch version is below 1.13)"
        raise ValueError(msg)

    x_arr: np.ndarray = x.cpu().numpy(**kwargs)
    if dtype is not None:  # supports older numpy version
        x_arr = x_arr.astype(dtype=dtype)  # type: ignore
    return x_arr


def numpy_to_tensor(
    x: Union[np.ndarray, np.number],
    *,
    device: DeviceLike = None,
    dtype: DTypeLike = None,
) -> Tensor:
    """Convert numpy array to PyTorch tensor."""
    device = get_device(device)
    dtype = get_dtype(dtype)
    return torch.from_numpy(x).to(dtype=dtype, device=device)


def numpy_view_as_real(x: np.ndarray) -> np.ndarray:
    """Convert complex array to float array.

    Args:
        x: The input complex array of any shape (...,)
    Returns:
        x_real: The same data in a float array of shape (..., 2)
    """
    assert numpy_is_complex(x)
    float_dtype = numpy_complex_dtype_to_float_dtype(x.dtype)
    if x.ndim > 0:
        return x.view(float_dtype).reshape(*x.shape, 2)
    else:
        # note: rebuild array here because view does not work on 0d arrays
        return np.array([x.real, x.imag], dtype=float_dtype)


def numpy_complex_dtype_to_float_dtype(dtype: np.dtype) -> np.dtype:
    """Returns the associated float dtype from complex dtype. If input dtype is not complex, it just returns the same dtype."""
    return np.empty((0,), dtype=dtype).real.dtype


def numpy_view_as_complex(x: np.ndarray) -> np.ndarray:
    """Convert complex array to float array.

    Args:
        x: The input float array of any shape (..., 2)
    Returns:
        x_real: The same data in a complex array of shape (...,)
    """
    assert not numpy_is_complex(x)
    return x[..., 0] + x[..., 1] * 1j


def numpy_is_floating_point(x: Union[np.ndarray, np.generic]) -> bool:
    return isinstance(x, np.floating)


def numpy_is_complex(x: Union[np.ndarray, np.generic]) -> bool:
    return np.iscomplexobj(x)


def numpy_is_complex_dtype(dtype: np.dtype) -> bool:
    return np.iscomplexobj(np.empty((0,), dtype=dtype))


def is_numpy_bool_array(x: Any) -> TypeIs[Union[np.bool_, np.ndarray]]:
    return isinstance(x, (np.generic, np.ndarray)) and x.dtype.kind == "b"


def is_numpy_number_like(x: Any) -> TypeIs[NumpyNumberLike]:
    """Returns True if x is an instance of a numpy number type, a np.bool_ or a zero-dimensional numpy array.
    If numpy is not installed, this function always returns False.
    """
    return isinstance(x, (np.number, np.bool_)) or (
        isinstance(x, np.ndarray) and x.ndim == 0
    )


def is_numpy_scalar_like(x: Any) -> TypeIs[NumpyScalarLike]:
    """Returns True if x is an instance of a numpy number type or a zero-dimensional numpy array.
    If numpy is not installed, this function always returns False.
    """
    return isinstance(x, np.generic) or (isinstance(x, np.ndarray) and x.ndim == 0)


def logical_and_lst(
    arr0: np.ndarray,
    /,
    *others: np.ndarray,
) -> np.ndarray:
    """Compute multiple logical_and over numpy arrays."""
    current = arr0
    for other in others:
        current = np.logical_and(current, other)
    return current


def logical_or_lst(
    arr0: np.ndarray,
    /,
    *others: np.ndarray,
) -> np.ndarray:
    """Compute multiple logical_or over numpy arrays."""
    current = arr0
    for other in others:
        current = np.logical_or(current, other)
    return current


def numpy_topk(
    x: np.ndarray,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    x_pt = numpy_to_tensor(x)
    values, indices = x_pt.topk(k=k, dim=dim, largest=largest, sorted=sorted)
    values = tensor_to_numpy(values)
    indices = tensor_to_numpy(indices)
    return values, indices


def numpy_item(x: Union[np.ndarray, np.generic, BuiltinScalar]) -> np.generic:
    if isinstance(x, np.generic):
        return x
    if is_builtin_scalar(x, strict=True):
        return np.array(x)[()]
    if prod(x.shape) != 1:
        msg = f"Invalid argument shape {x.shape=}. (expected nd-array with 1 element)"
        raise ValueError(msg)

    indices = tuple([0] * x.ndim)
    return x[indices]


@overload
def numpy_all_eq(
    x: Union[np.generic, np.ndarray],
    dim: None = None,
) -> bool:
    ...


@overload
def numpy_all_eq(
    x: Union[np.generic, np.ndarray],
    dim: int,
) -> np.ndarray:
    ...


def numpy_all_eq(
    x: Union[np.generic, np.ndarray],
    dim: Union[int, None] = None,
) -> bool:
    if dim is None:
        if x.ndim == 0 or prod(x.shape) == 0:
            return True
        else:
            return (x.flat[0] == x.flat[1:]).all()

    else:
        slices = [slice(None) for _ in range(x.ndim)]
        slices[dim] = 0
        slices.insert(dim + 1, None)
        slices = tuple(slices)
        return (x == x[slices]).all(dim)


def numpy_all_ne(x: Union[np.generic, np.ndarray]) -> bool:
    return len(np.unique(x)) == x.size



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pickle
import random
import tempfile
import time
import unittest
from pathlib import Path
from unittest import TestCase

import numpy as np
import torch
from torch import Tensor
from torch.utils.data.dataset import Subset
from torchvision.datasets import CIFAR10

from torchoutil.extras.hdf import HDFDataset, pack_to_hdf
from torchoutil.nn import ESequential, IndexToOnehot, ToList, ToNumpy
from torchoutil.nn.functional import to_tensor
from torchoutil.pyoutil import dict_list_to_list_dict


class TestHDF(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

    def test_cifar10_pack_to_hdf(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        dataset = CIFAR10(
            tmpdir,
            train=False,
            transform=ToNumpy(),
            target_transform=ESequential(IndexToOnehot(10), ToList()),
            download=True,
        )
        indices = torch.randint(0, len(dataset), (max(len(dataset) // 10, 1),)).tolist()
        dataset = Subset(dataset, indices)

        path = tmpdir.joinpath("test_cifar10.hdf")
        hdf_dataset = pack_to_hdf(dataset, path, exists="overwrite")

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = hdf_dataset[idx]

        assert len(dataset) == len(hdf_dataset)
        assert np.equal(image0, image1).all()
        assert np.equal(label0, label1).all()

        # Try pickle for DDP support
        dumped = pickle.dumps(hdf_dataset)
        hdf_dataset_2 = pickle.loads(dumped)
        assert hdf_dataset == hdf_dataset_2

        hdf_dataset_2.close(remove_file=False)
        hdf_dataset.close(remove_file=True)

    def test_shape_column(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        shape_suffix = "_added_shape"
        data = [torch.rand(10, 2), torch.rand(10, 5), torch.rand(10, 3)]
        data_shape = [(10, 1), (10, 2), (10, 1)]

        ds_dict = {"data": data, f"data{shape_suffix}": data_shape}
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_shape.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            shape_suffix=shape_suffix,
            ds_kwds=dict(cast="to_torch_or_builtin"),
        )

        assert len(hdf_dataset.added_columns) == 0
        assert len(hdf_dataset) == len(ds_list)

        for i, item in enumerate(iter(hdf_dataset)):
            assert set(item.keys()) == set(ds_dict.keys())

            data_i = item["data"]
            assert isinstance(data_i, Tensor)
            assert (
                data_i.shape == data_shape[i]
            ), f"{i=}; {data_i.shape=}; {data_shape[i]=}"

        hdf_dataset.close(remove_file=True)

    def test_indices_mask(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        data = np.random.rand(100, 8)
        ds_dict = {"data": data}
        ds_list = dict_list_to_list_dict(ds_dict)

        path = tmpdir.joinpath("test_indices_slicing.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            ds_kwds=dict(cast="none"),
        )

        # Random indices test
        # with generate indices with duplicates in random order !
        indices = np.random.randint(0, len(data), (10,))
        assert (hdf_dataset[indices, "data"] == data[indices]).all()

        # Random mask test
        mask = np.random.rand(len(data)) > 0.2
        indices = np.where(mask)[0]

        expected = hdf_dataset[mask, "data"]
        for values in (
            hdf_dataset[indices, "data"],
            hdf_dataset[mask.tolist(), "data"],
            hdf_dataset[indices.tolist(), "data"],
            hdf_dataset[to_tensor(mask), "data"],
            hdf_dataset[to_tensor(indices), "data"],
            data[mask],
            data[indices],
        ):
            assert (expected == values).all()

        hdf_dataset.close(remove_file=True)

    def test_special_floating_dtypes(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "f16": torch.rand(10, 2, 3, dtype=torch.float16),
            "f32": torch.rand(10, 2, 3, dtype=torch.float32),
            "f64": torch.rand(10, 2, 3, dtype=torch.float64),
            "c64": torch.rand(10, 2, 3, dtype=torch.complex64),
            "c128": torch.rand(10, 2, 3, dtype=torch.complex128),
            "bool": torch.rand(10, 1) > 0.5,
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")
        keys = set(ds_dict.keys())

        fname = "test_special_floating_dtypes.hdf"
        fpath = tmpdir.joinpath(fname)
        hdf_ds = pack_to_hdf(
            ds_list,
            fpath,
            exists="overwrite",
            ds_kwds=dict(cast="to_torch_src"),
        )

        assert len(hdf_ds.added_columns) == 0
        assert len(hdf_ds) == len(ds_list)

        for i, hdf_item in enumerate(iter(hdf_ds)):
            assert set(hdf_item.keys()) == keys
            assert set(ds_list[i].keys()) == keys

            src_item = ds_list[i]
            for k in keys:
                msg = f"Index: {i}; Key: {k}; {src_item[k]=}; {hdf_item[k]=}"
                src_value = src_item[k]
                hdf_value = hdf_item[k]
                assert torch.equal(src_value, hdf_value), msg

        hdf_ds.close(remove_file=True)

    def test_slice(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "a": list(range(10, 20)),
            "b": torch.rand(10, 2, 3),
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_slice.hdf")
        pack_to_hdf(ds_list, path, exists="overwrite", ds_kwds=dict(open_hdf=False))

        hdf_dataset = HDFDataset(path, cast="to_builtin")

        assert len(hdf_dataset) == len(ds_list)
        assert hdf_dataset[:, "a"] == ds_dict["a"]
        assert hdf_dataset[:, "b"] == ds_dict["b"].tolist()

        indices = torch.randperm(len(hdf_dataset))

        assert hdf_dataset[indices, "a"] == [ds_dict["a"][idx] for idx in indices]
        assert (hdf_dataset[indices, "b"] == ds_dict["b"][indices].numpy()).all()

    def test_string(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "int": list(range(10)),
            "string": [
                "".join(map(str, range(random.randint(10, 100)))) for _ in range(10)
            ],
            "list_string": [
                [],
                ["aa", "bbb"],
                [],
                ["cccc"],
                [""],
                ["dd", "", "e", "ff", "gggggg"],
                ["éééé", "û", "é"],
            ]
            + [[]] * 3,
            "empty_lists": [[]] * 10,
            "bytes": [b""] * 6 + [b"a2", b"dnqzu1dhqz", b"0djqizjdz", b"du12qzd"],
            # TODO: add support for bytearray ?
            "bytearray": np.array([bytearray(str(i), "utf-8") for i in range(10, 20)]),
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_string.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            ds_kwds=dict(cast="to_builtin"),
        )

        assert hdf_dataset._src_is_unicode == {
            "int": False,
            "string": True,
            "list_string": True,
            "empty_lists": False,
            "bytes": False,
            "bytearray": False,
        }
        idx = torch.randint(0, len(hdf_dataset), ()).item()
        col = random.choice(list(ds_dict.keys()))
        eq = hdf_dataset[idx, col] == ds_dict[col][idx]
        if isinstance(eq, np.ndarray):
            eq = eq.all()
        assert eq, f"{idx=}; {col=}"

        assert len(hdf_dataset) == len(ds_list)
        for k in ds_dict.keys():
            hdf_col = hdf_dataset[:, k]
            ds_col = ds_dict[k]
            eq = hdf_col == ds_col
            if isinstance(eq, np.ndarray):
                eq = eq.all()
            assert eq, f"{k=}, {hdf_col=} != {ds_col=}"

    def test_string_comp(self) -> None:
        num_data = 10000
        max_string_len = 100
        max_sublst = 5
        seed = 42

        gen = torch.Generator().manual_seed(seed)
        # note: vlen_str does not support chr(0)
        strings = [
            "".join(
                map(
                    chr,
                    torch.randint(1, 256, (max_string_len,), generator=gen).tolist(),
                )
            )
            for _ in range(num_data)
        ]
        list_strings = [
            [
                "".join(
                    map(
                        chr,
                        torch.randint(
                            1, 256, (max_string_len,), generator=gen
                        ).tolist(),
                    )
                )
                for _ in range(1, max_sublst + 1)
            ]
            for _ in range(num_data)
        ]
        ds_dict = {"strings": strings, "list_strings": list_strings}
        ds_list = dict_list_to_list_dict(ds_dict, key_mode="same")

        ds_vlen = pack_to_hdf(
            ds_list,
            self.tmpdir.joinpath("test_string_comp_vlen.hdf"),
            exists="overwrite",
            store_str_as_vlen=True,
            verbose=2,
        )
        ds_bytes = pack_to_hdf(
            ds_list,
            self.tmpdir.joinpath("test_string_comp_bytes.hdf"),
            exists="overwrite",
            store_str_as_vlen=False,
            verbose=2,
        )

        n_try = 10
        duration_ds_vlen_lst = []
        duration_ds_bytes_lst = []

        for _ in range(n_try):
            start = time.perf_counter()
            ds_bytes_data = ds_bytes[:]
            duration_ds_bytes = time.perf_counter() - start

            start = time.perf_counter()
            ds_vlen_data = ds_vlen[:]
            duration_ds_vlen = time.perf_counter() - start

            duration_ds_bytes_lst.append(duration_ds_bytes)
            duration_ds_vlen_lst.append(duration_ds_vlen)
            assert ds_bytes_data.keys() == ds_vlen_data.keys()
            assert all(
                np.all(ds_bytes_data[k] == ds_vlen_data[k])
                for k in ds_bytes_data.keys()
            )

        assert np.median(duration_ds_bytes_lst) < np.median(duration_ds_vlen_lst)

        ds_vlen.close(remove_file=True)
        ds_bytes.close(remove_file=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, List, Union

import torch
from torch import Tensor
from torch.types import Number

from torchoutil.core.get import CUDA_IF_AVAILABLE, DeviceLike, get_device, get_generator
from torchoutil.types import LongTensor1D, Tensor1D, is_builtin_number


def get_inverse_perm(indices: Tensor, dim: int = -1) -> Tensor:
    """Return inverse permutation indices.
    The output will be a tensor of shape (..., N).

    Args:
        indices: Original permutation indices as tensor of shape (..., N).
        dim: Dimension of indices. defaults to -1.

    Example 1
    ----------
        >>> x = torch.as_tensor([2, 4, 8, 10])
        >>> indices = torch.randperm(len(x))
        >>> x = x[indices]
        >>> # x is now shuffled, to get back the original order we need the indices
        >>> inv_indices = get_inverse_perm(indices)
        >>> x_reordered = x[inv_indices]
        >>> x_reordered
        ... tensor([2, 4, 8, 10])
    """
    arange = torch.arange(
        indices.shape[dim],
        dtype=indices.dtype,
        device=indices.device,
    )
    arange = arange.expand(*indices.shape)
    indices_inv = torch.empty_like(indices)
    indices_inv = indices_inv.scatter(dim, indices, arange)
    return indices_inv


def randperm_diff(
    size: int,
    generator: Union[None, int, torch.Generator] = None,
    device: DeviceLike = CUDA_IF_AVAILABLE,
) -> LongTensor1D:
    """This function ensure that every value i cannot be the element at index i.
    The output will be a tensor of shape (size,).

    Args:
        size: The number of indices. Cannot be < 2.
        seed: The seed or torch.Generator used to generate permutation.
        device: The PyTorch device of the output indices tensor.

    Example 1
    ----------
        >>> torch.randperm(5)
        tensor([1, 4, 2, 5, 0])  # 2 is the element of index 2 !
        >>> randperm_diff(5)
        tensor([2, 0, 4, 1, 3])
    """
    if size < 2:
        raise ValueError(f"Invalid argument {size=} < 2 for randperm_diff.")

    device = get_device(device)
    generator = get_generator(generator)

    perm_kws: Dict[str, Any] = dict(generator=generator, device=device)
    arange = torch.arange(size, device=device)
    perm = torch.randperm(size, **perm_kws)

    while perm.eq(arange).any():
        perm = torch.randperm(size, **perm_kws)
    return perm  # type: ignore


def get_perm_indices(x1: Tensor, x2: Tensor) -> LongTensor1D:
    """Find permutation between two vectors t1 and t2 which contains values from 0 to N-1.

    Example 1::
    -----------
        >>> x1 = torch.as_tensor([0, 1, 2, 4, 3, 6, 5, 7])
        >>> x2 = torch.as_tensor([0, 2, 1, 4, 3, 5, 6, 7])
        >>> indices = get_perm_indices(x1, x2)
        >>> torch.equal(x1, x2[indices])
        True
    """
    indices = (x1[..., None, :] == x2[..., :, None]).short().argmax(dim=-2)
    return indices  # type: ignore


def insert_at_indices(
    x: Tensor,
    indices: Union[Tensor, List, Number],
    values: Union[Number, Tensor],
) -> Tensor1D:
    """Insert value(s) in vector at specified indices.

    Example 1::
    -----------
        >>> x = torch.as_tensor([1, 1, 2, 2, 2, 3])
        >>> indices = torch.as_tensor([2, 5])
        >>> values = 4
        >>> insert_values(x, indices, values)
        tensor([1, 1, 4, 2, 2, 2, 4, 3])
    """
    if x.ndim != 1:
        msg = f"Invalid argument number of dims. (found {x.ndim=} but expected 1)"
        raise ValueError(msg)

    device = x.device
    if isinstance(indices, Tensor):
        pass
    elif is_builtin_number(indices):
        indices = torch.as_tensor([indices], device=device, dtype=torch.long)
    elif isinstance(indices, list):
        indices = torch.as_tensor(indices, device=device, dtype=torch.long)
    else:
        raise TypeError(f"Invalid argument type {type(indices)=}.")

    out = torch.empty((x.shape[0] + indices.shape[0]), dtype=x.dtype, device=device)
    indices = indices + torch.arange(
        indices.shape[0], device=indices.device, dtype=indices.dtype
    )
    out[indices] = values
    mask = torch.full((out.shape[0],), True, dtype=torch.bool)
    mask[indices] = False
    out[mask] = x
    return out


def remove_at_indices(
    x: Tensor,
    indices: Union[Tensor, List, Number],
) -> Tensor1D:
    """Remove value(s) in vector at specified indices."""
    if x.ndim != 1:
        msg = f"Invalid argument number of dims. (found {x.ndim=} but expected 1)"
        raise ValueError(msg)

    device = x.device
    if isinstance(indices, Tensor):
        pass
    elif is_builtin_number(indices):
        indices = torch.as_tensor([indices], device=device, dtype=torch.long)
    elif isinstance(indices, list):
        indices = torch.as_tensor(indices, device=device, dtype=torch.long)
    else:
        raise TypeError(f"Invalid argument type {type(indices)=}.")

    indices = indices + torch.arange(
        indices.shape[0], device=device, dtype=indices.dtype
    )
    mask = torch.full((x.shape[0],), True, dtype=torch.bool)
    mask[indices] = False
    out = x[mask]
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import math
import shutil
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union

import torch
from torch import Tensor
from torch.utils.data.dataloader import DataLoader

import pyoutil as po
import torchoutil as to
from torchoutil import nn
from torchoutil.core.packaging import _NUMPY_AVAILABLE, _TQDM_AVAILABLE
from torchoutil.extras.numpy.scan_info import (
    merge_numpy_dtypes,
    numpy_dtype_to_fill_value,
    scan_shape_dtypes,
)
from torchoutil.utils.data.dataloader import get_auto_num_cpus
from torchoutil.utils.data.dataset import (
    SizedDatasetLike,
    SizedIterableDatasetLike,
    TransformWrapper,
)
from torchoutil.utils.pack.common import (
    ATTRS_FNAME,
    CONTENT_DNAME,
    EXISTS_MODES,
    ContentMode,
    ExistsMode,
    _tuple_to_dict,
)
from torchoutil.utils.pack.dataset import PackedDataset
from torchoutil.utils.saving.common import to_builtin
from torchoutil.utils.saving.save_fn import SAVE_EXTENSIONS, SAVE_FNS, SaveFnLike

if _TQDM_AVAILABLE:
    import tqdm
if _NUMPY_AVAILABLE:
    import numpy as np


T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)
T_DictOrTuple = TypeVar("T_DictOrTuple", tuple, dict, covariant=True)

pylog = logging.getLogger(__name__)


@torch.inference_mode()
def pack_dataset(
    dataset: SizedDatasetLike[T],
    root: Union[str, Path],
    pre_transform: Optional[Callable[[T], U]] = None,
    *,
    batch_size: int = 32,
    content_mode: ContentMode = "item",
    custom_file_fmt: Union[None, str, Callable[[int], str]] = None,
    ds_kwds: Optional[Dict[str, Any]] = None,
    exists: ExistsMode = "error",
    save_fn: SaveFnLike[Union[U, List[U]]] = "pickle",
    subdir_size: Optional[int] = 100,
    num_workers: Union[int, Literal["auto"]] = "auto",
    transform_in_worker: bool = True,
    user_attrs: Optional[Dict[str, Any]] = None,
    verbose: int = 0,
) -> PackedDataset[U, U]:
    """Pack a dataset to pickle files.

    Here is an example how files are stored on disk for a dataset containing 1000 items:

    .. code-block:: text
        :caption:  Dataset folder tree example

        {root}
        ├── attributes.json
        └── data
            ├── 0
            |   └── 100 pt files
            ├── 1
            |   └── 100 pt files
            ├── ...
            |   └── ...
            └── 9
                └── 100 pt files

    Args:
        dataset: Dataset-like to pack.
        root: Directory to store pickled data.
        pre_transform: Transform to apply to each item before saving. defaults to None.
        batch_size: Batch size used by the dataloader. defaults to 32.
        num_workers: Number of workers used by the dataloader. defaults to "auto".
        exists: Determine which action should be performed if the target HDF file already exists.
            "overwrite": Replace the target files then pack dataset.
            "skip": Skip this function and returns the packed dataset.
            "error": Raises a ValueError.
        content_mode: Specify how the data should be stored.
            If "item", each dataset item will be stored in a separate file.
            If "batch", each dataset item will be stored in a batch file of size batch_size.
        custom_file_fmt: Custom file format.
            If None, defaults to "{{i:0{num_digits}d}}.pt".
            defaults to None.
        save_fn: Custom save function to save an item or a batch. defaults to torch.save.
        subdir_size: Optional number of files per folder.
            Using None will disable subdir an put all files in data/ folder.
            defaults to 100.
        transform_in_worker: If True, appoly transform in parallel with workers' Dataloader.
        verbose: Verbose level during packing. Higher value means more messages. defaults to 0.
    """
    if content_mode == "column":
        if custom_file_fmt is None:
            custom_file_fmt = "{column}.{ext}"
        if not isinstance(custom_file_fmt, str):
            msg = (
                f"Invalid argument type {type(custom_file_fmt)=} with {content_mode=}."
            )
            raise ValueError(msg)

        return pack_dataset_to_columns(
            dataset=dataset,
            root=root,
            pre_transform=pre_transform,  # type: ignore
            batch_size=batch_size,
            ds_kwds=ds_kwds,
            exists=exists,
            fname_fmt=custom_file_fmt,
            num_workers=num_workers,
            save_fn=save_fn,  # type: ignore
            user_attrs=user_attrs,
            verbose=verbose,
        )

    # Check inputs
    if not isinstance(dataset, SizedDatasetLike):
        msg = f"Cannot pack to hdf a non-sized-dataset '{dataset.__class__.__name__}'."
        raise TypeError(msg)
    if len(dataset) == 0:
        raise ValueError("Cannot pack to hdf an empty dataset.")
    if transform_in_worker and isinstance(dataset, SizedIterableDatasetLike):
        msg = "Cannot apply transform in worker with an iterable dataset kind."
        raise NotImplementedError(msg)

    if num_workers == "auto":
        num_workers = get_auto_num_cpus()

    if isinstance(save_fn, str):
        ext = SAVE_EXTENSIONS[save_fn]
        save_fn = SAVE_FNS[save_fn]
    else:
        ext = "bin"

    packed, root, content_dpath = _setup_args(root, exists)
    if packed is not None:
        return packed

    if transform_in_worker:
        wrapped = TransformWrapper(dataset, pre_transform)
        src_dataset_name = wrapped.unwrap().__class__.__name__
    else:
        wrapped = dataset
        src_dataset_name = dataset.__class__.__name__

    loader = DataLoader(
        wrapped,  # type: ignore
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=nn.Identity(),
        drop_last=False,
        pin_memory=False,
    )

    if content_mode == "item":
        num_files = len(wrapped)
    elif content_mode == "batch":
        num_files = math.ceil(len(wrapped) / batch_size)
    else:
        raise ValueError(f"Invalid argument {content_mode=}.")

    if custom_file_fmt is None:
        num_digits = math.ceil(math.log10(num_files))
        file_fmt = f"{{:0{num_digits}d}}.{ext}".format
    elif isinstance(custom_file_fmt, str):
        file_fmt = custom_file_fmt.format
    else:
        file_fmt = custom_file_fmt

    fnames = [file_fmt(i) for i in range(num_files)]

    if subdir_size is not None and len(fnames) > subdir_size:
        num_subdirs = math.ceil(len(fnames) / subdir_size)
        num_digits = math.ceil(math.log10(num_subdirs))
        dir_fmt = f"{{:0{num_digits}d}}".format
        sub_dnames = [dir_fmt(i) for i in range(num_subdirs)]
        fnames = [
            str(Path(sub_dnames[i // subdir_size]).joinpath(fname))
            for i, fname in enumerate(fnames)
        ]

    fpaths = [content_dpath.joinpath(fname) for fname in fnames]

    if custom_file_fmt is not None or subdir_size is not None:
        unique_parents = dict.fromkeys(fpath.parent for fpath in fpaths)
        for parent in unique_parents:
            parent.mkdir(parents=True, exist_ok=True)

    if _TQDM_AVAILABLE and verbose >= 1:
        loader = tqdm.tqdm(
            loader,
            total=len(loader),
            desc=f"Packing {len(fnames)} items...",
        )

    i = 0
    for batch_lst in loader:
        if pre_transform is not None and not transform_in_worker:
            batch_lst = [pre_transform(item) for item in batch_lst]

        if content_mode == "item":
            for item in batch_lst:
                fpath = fpaths[i]
                save_fn(item, fpath)
                i += 1

        elif content_mode == "batch":
            fpath = fpaths[i]
            save_fn(batch_lst, fpath)
            i += 1

        else:
            raise ValueError(f"Invalid argument {content_mode=}.")

    if hasattr(dataset, "info"):
        info = dataset.info  # type: ignore
        info = to_builtin(info)
    else:
        info = {}

    if hasattr(dataset, "attrs"):
        source_attrs = dataset.attrs  # type: ignore
        source_attrs = to_builtin(source_attrs)
    else:
        source_attrs = {}

    attributes = {
        "source_dataset": src_dataset_name,
        "length": len(wrapped),
        "creation_date": po.now_iso(),
        "batch_size": batch_size,
        "content_mode": content_mode,
        "content_dname": CONTENT_DNAME,
        "info": info,
        "source_attrs": source_attrs,
        "num_files": len(fnames),
        "files": fnames,
        "subdir_size": subdir_size,
        "item_type": "raw",
        "column_to_fname": {},
        "user_attrs": to_builtin(user_attrs),
    }
    attrs_fpath = root.joinpath(ATTRS_FNAME)
    with open(attrs_fpath, "w") as file:
        json.dump(attributes, file, indent="\t")

    if ds_kwds is None:
        ds_kwds = {}
    packed = PackedDataset(root, **ds_kwds)
    return packed


def pack_dataset_to_columns(
    dataset: SizedDatasetLike[T],
    root: Union[str, Path],
    pre_transform: Optional[Callable[[T], T_DictOrTuple]] = None,
    *,
    batch_size: int = 32,
    ds_kwds: Optional[Dict[str, Any]] = None,
    exists: ExistsMode = "error",
    fname_fmt: str = "{column}.{ext}",
    num_workers: Union[int, Literal["auto"]] = "auto",
    save_fn: SaveFnLike[np.ndarray] = "pickle",
    user_attrs: Optional[Dict[str, Any]] = None,
    verbose: int = 0,
) -> PackedDataset[T_DictOrTuple, T_DictOrTuple]:
    if not isinstance(dataset, SizedDatasetLike):
        msg = f"Cannot pack to hdf a non-sized-dataset '{dataset.__class__.__name__}'."
        raise TypeError(msg)
    if len(dataset) == 0:
        raise ValueError("Cannot pack to hdf an empty dataset.")

    if not _NUMPY_AVAILABLE:
        msg = f"Function '{po.get_current_fn_name()}' cannot be called without numpy installed. Please install it with `pip install torchoutil[extras]`."
        raise RuntimeError(msg)

    packed, root, content_dpath = _setup_args(root, exists)
    if packed is not None:
        return packed

    if pre_transform is None:
        pre_transform = nn.Identity()
    if num_workers == "auto":
        num_workers = get_auto_num_cpus()

    src_dataset_name = dataset.__class__.__name__

    item_0 = next(iter(dataset))
    item_0 = pre_transform(item_0)

    dict_pre_transform: Callable[[T], Dict[str, Any]]

    if po.is_dict_str(item_0):
        item_type = "dict"
        dict_pre_transform = pre_transform  # type: ignore

    elif isinstance(item_0, tuple):
        item_type = "tuple"
        dict_pre_transform = po.compose(pre_transform, _tuple_to_dict)
        item_0 = _tuple_to_dict(item_0)

    else:
        msg = f"Invalid item type for {dataset.__class__.__name__}. (expected Dict[str, Any] or tuple but found {type(item_0)})"
        raise ValueError(msg)

    infos_0 = {name: scan_shape_dtypes(value) for name, value in item_0.items()}
    max_shapes = {name: info.shape for name, info in infos_0.items()}
    dtypes = {name: info.numpy_dtype for name, info in infos_0.items()}

    del pre_transform, item_0, infos_0

    loader = DataLoader(
        dataset,  # type: ignore
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=nn.Identity(),
        drop_last=False,
        pin_memory=False,
    )

    if _TQDM_AVAILABLE:
        loader = tqdm.tqdm(loader, total=len(loader), disable=verbose < 1)

    i = 0
    for batch in loader:
        batch = [dict_pre_transform(item) for item in batch]
        for item in batch:
            for name, value in item.items():
                info = scan_shape_dtypes(value)

                if info.ndim != len(max_shapes[name]):
                    msg = f"Invalid argument shape at index {i} with key {name}. (found {info.ndim} ndim but expected {len(max_shapes[name])})"
                    raise ValueError(msg)

                max_shapes[name] = tuple(
                    max(info.shape[i], max_shapes[name][i]) for i in range(info.ndim)
                )
                dtypes[name] = merge_numpy_dtypes([info.numpy_dtype, dtypes[name]])
            i += 1

    fill_values = {
        name: numpy_dtype_to_fill_value(np_dtype) for name, np_dtype in dtypes.items()
    }
    data_dict = {
        name: np.full(
            (len(dataset),) + shape,
            fill_value=fill_values[name],
            dtype=dtypes[name],
        )
        for name, shape in max_shapes.items()
    }

    i = 0
    it = iter(loader)
    if _TQDM_AVAILABLE:
        it = tqdm.tqdm(it, total=len(loader), disable=verbose < 1)

    for batch in it:
        batch = [dict_pre_transform(item) for item in batch]
        for item in batch:
            for k, v in item.items():
                v = to.to_numpy(v)
                slices = [slice(dim_i) for dim_i in v.shape]
                slices = (i,) + tuple(slices)
                data_dict[k][slices] = v
            i += 1

    if hasattr(dataset, "info"):
        info = dataset.info  # type: ignore
        info = to_builtin(info)
    else:
        info = {}

    if hasattr(dataset, "attrs"):
        source_attrs = dataset.attrs  # type: ignore
        source_attrs = to_builtin(source_attrs)
    else:
        source_attrs = {}

    src_attrs = {
        "source_dataset": src_dataset_name,
        "batch_size": len(dataset),
        "item_type": item_type,
        "info": info,
        "source_attrs": source_attrs,
        "subdir_size": None,
        "user_attrs": user_attrs,
    }
    return _pack_dataset_dict(
        data_dict=data_dict,  # type: ignore
        root=root,
        content_dpath=content_dpath,
        src_attrs=src_attrs,
        save_fn=save_fn,
        ds_kwds=ds_kwds,
        fname_fmt=fname_fmt,
        verbose=verbose,
    )


def pack_dataset_dict(
    data_dict: Dict[str, Union[np.ndarray, Tensor]],
    root: Union[str, Path],
    *,
    ds_kwds: Optional[Dict[str, Any]] = None,
    exists: ExistsMode = "error",
    fname_fmt: str = "{column}.{ext}",
    save_fn: SaveFnLike[np.ndarray] = "pickle",
    user_attrs: Optional[Dict[str, Any]] = None,
    verbose: int = 0,
) -> PackedDataset:
    if not po.all_eq(map(len, data_dict.values())):
        raise ValueError

    packed, root, content_dpath = _setup_args(root, exists)
    if packed is not None:
        return packed

    src_attrs = {
        "source_dataset": data_dict.__class__.__name__,
        "batch_size": len(next(iter(data_dict.values()))),
        "item_type": "dict",
        "info": {},
        "source_attrs": {},
        "subdir_size": None,
        "user_attrs": to_builtin(user_attrs),
    }
    return _pack_dataset_dict(
        data_dict=data_dict,
        root=root,
        content_dpath=content_dpath,
        ds_kwds=ds_kwds,
        fname_fmt=fname_fmt,
        save_fn=save_fn,
        src_attrs=src_attrs,
        verbose=verbose,
    )


def _pack_dataset_dict(
    data_dict: Dict[str, Any],
    root: Path,
    content_dpath: Path,
    ds_kwds: Optional[Dict[str, Any]] = None,
    fname_fmt: str = "{column}.{ext}",
    save_fn: SaveFnLike[np.ndarray] = "pickle",
    src_attrs: Optional[Dict[str, Any]] = None,
    verbose: int = 0,
) -> PackedDataset:
    if isinstance(save_fn, str):
        ext = SAVE_EXTENSIONS[save_fn]
        save_fn = SAVE_FNS[save_fn]
    else:
        ext = "bin"

    data_dict = {k: to.to_numpy(v) for k, v in data_dict.items()}

    column_to_fname = {}
    for column, values in tqdm.tqdm(data_dict.items(), disable=verbose < 1):
        fname = fname_fmt.format(column=column, ext=ext)
        column_to_fname[column] = fname
        fpath = content_dpath.joinpath(fname)
        save_fn(values, fpath)

    fnames = list(dict.fromkeys(column_to_fname.values()))
    attrs = {
        "length": len(next(iter(data_dict.values()))),
        "creation_date": po.now_iso(),
        "content_mode": "column",
        "content_dname": CONTENT_DNAME,
        "num_files": len(fnames),
        "files": fnames,
        "column_to_fname": column_to_fname,
    }
    if src_attrs is not None:
        attrs.update(src_attrs)

    attrs_fpath = root.joinpath(ATTRS_FNAME)
    with open(attrs_fpath, "w") as file:
        json.dump(attrs, file, indent="\t")

    if ds_kwds is None:
        ds_kwds = {}
    packed = PackedDataset(root, **ds_kwds)
    return packed


def _setup_args(
    root: Union[str, Path],
    exists: ExistsMode,
) -> Tuple[Optional[PackedDataset], Path, Path]:
    root = Path(root).resolve()
    if root.exists() and not root.is_dir():
        raise RuntimeError(f"Path {root=} exists but it is not a directory.")

    content_dpath = root.joinpath(CONTENT_DNAME)

    if not content_dpath.is_dir():
        pass
    elif exists == "error":
        msg = f"Cannot overwrite root data {str(content_dpath)}. Please remove it or use exists='overwrite' option."
        raise ValueError(msg)
    elif exists == "skip":
        return PackedDataset(root), root, content_dpath
    elif exists == "overwrite":
        shutil.rmtree(content_dpath)
    else:
        msg = f"Invalid argument {exists=}. (expected one of {EXISTS_MODES})"
        raise ValueError(msg)

    content_dpath.mkdir(parents=True, exist_ok=True)
    return None, root, content_dpath



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import (
    np,
    numpy_is_complex,
    numpy_is_complex_dtype,
    numpy_item,
    numpy_to_tensor,
    numpy_view_as_complex,
    numpy_view_as_real,
    tensor_to_numpy,
)


class TestNumpyConversions(TestCase):
    def test_example_1(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        x_tensor = torch.rand(3, 4, 5)
        x_array = tensor_to_numpy(x_tensor)
        result = numpy_to_tensor(x_array)

        assert torch.equal(x_tensor, result)

    def test_complex(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        complex_dtypes = [np.complex64, np.complex128]
        x_complex = [
            np.array(
                np.random.rand(1) * 1j,
                dtype=complex_dtypes[np.random.randint(0, len(complex_dtypes))],
            )
            for _ in range(1000)
        ]
        assert all(numpy_is_complex(xi) for xi in x_complex)
        assert all(numpy_is_complex_dtype(xi.dtype) for xi in x_complex)

        x_real = [numpy_view_as_real(xi) for xi in x_complex]
        assert all(not numpy_is_complex(xi) for xi in x_real)
        assert all(not numpy_is_complex_dtype(xi.dtype) for xi in x_real)

        result = [numpy_view_as_complex(xi) for xi in x_real]
        assert all(numpy_is_complex(xi) for xi in result)
        assert all(numpy_is_complex_dtype(xi.dtype) for xi in result)
        assert x_complex == result

    def test_numpy_item(self) -> None:
        x = np.array(1.2)
        result = numpy_item(x)
        assert isinstance(result, np.generic)

        x = np.array([1, 2])[1]
        result = numpy_item(x)
        assert isinstance(result, np.generic)

        x = 10.99
        result = numpy_item(x)
        assert isinstance(result, np.generic)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import math

from torch.optim.lr_scheduler import LambdaLR
from torch.optim.optimizer import Optimizer

pylog = logging.getLogger(__name__)


class CosDecayScheduler(LambdaLR):
    def __init__(
        self,
        optimizer: Optimizer,
        n_steps: int,
        last_epoch: int = -1,
    ) -> None:
        super().__init__(
            optimizer=optimizer,
            lr_lambda=CosDecayRule(n_steps),
            last_epoch=last_epoch,
        )


class CosDecayRule:
    # Note : use class instead of function for scheduler rules to become pickable for multiple-GPU with Lightning
    def __init__(self, n_steps: int) -> None:
        if n_steps < 0:
            msg = f"Invalid argument {n_steps=} < 0 in {self.__class__.__name__}."
            raise ValueError(msg)
        elif n_steps == 0:
            pylog.warning(
                f"Replacing {n_steps=} by n_steps=1 in {self.__class__.__name__}."
            )
            n_steps = max(n_steps, 1)

        super().__init__()
        self.n_steps = n_steps

    def __call__(self, step: int) -> float:
        step = min(step, self.n_steps - 1)
        return 0.5 * (1.0 + math.cos(math.pi * step / self.n_steps))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, List, Tuple, TypeVar, Union, final, overload

import torch
from torch import Tensor
from torch.utils.data.dataset import Dataset

from torchoutil.extras.numpy.functional import is_numpy_bool_array
from torchoutil.pyoutil.typing import is_iterable_bool, is_iterable_integral
from torchoutil.pyoutil.typing.classes import SupportsLenAndGetItem
from torchoutil.types import is_bool_tensor1d, is_integral_tensor1d, is_number_like
from torchoutil.types._typing import BoolTensor, Tensor1D
from torchoutil.utils.data.dataset import Wrapper

T = TypeVar("T", covariant=False)
U = TypeVar("U", covariant=False)

Indices = Union[Iterable[bool], Iterable[int], None, slice, Tensor1D]


class DatasetSlicer(Generic[T], ABC, Dataset[T]):
    def __init__(
        self,
        *,
        add_slice_support: bool = True,
        add_indices_support: bool = True,
        add_mask_support: bool = True,
        add_none_support: bool = True,
    ) -> None:
        Dataset.__init__(self)
        self._add_slice_support = add_slice_support
        self._add_indices_support = add_indices_support
        self._add_mask_support = add_mask_support
        self._add_none_support = add_none_support

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_item(self, idx, /, *args, **kwargs) -> Any:
        raise NotImplementedError

    @overload
    @final
    def __getitem__(self, idx: int, /) -> T:
        ...

    @overload
    @final
    def __getitem__(self, idx: Indices, /) -> List[T]:
        ...

    @overload
    @final
    def __getitem__(self, idx: Tuple[Any, ...], /) -> Any:
        ...

    @final
    def __getitem__(self, idx) -> Any:
        if isinstance(idx, tuple) and len(idx) > 1:
            idx, *args = idx
        else:
            args = ()

        if is_number_like(idx):
            return self.get_item(idx, *args)

        elif isinstance(idx, slice):
            return self.get_items_slice(idx, *args)

        elif (
            is_iterable_bool(idx)
            or is_bool_tensor1d(idx)
            or (is_numpy_bool_array(idx) and idx.ndim == 1)
        ):
            return self.get_items_mask(idx, *args)

        elif is_iterable_integral(idx) or is_integral_tensor1d(idx):
            return self.get_items_indices(idx, *args)

        elif idx is None:
            return self.get_items_none(idx, *args)

        else:
            raise TypeError(f"Invalid argument type {type(idx)=} with {args=}.")

    @final
    def __getitems__(
        self,
        indices: Indices,
        *args,
    ) -> List[T]:
        return self.__getitem__(indices, *args)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def get_items_indices(
        self,
        indices: Union[Iterable[int], Tensor1D],
        *args,
    ) -> List[T]:
        if self._add_indices_support:
            return [self.get_item(idx, *args) for idx in indices]
        else:
            return self.get_item(indices, *args)

    def get_items_mask(
        self,
        mask: Union[Iterable[bool], Tensor],
        *args,
    ) -> List[T]:
        if self._add_mask_support:
            if not isinstance(mask, Tensor):
                mask = torch.as_tensor(list(mask), dtype=torch.bool)  # type: ignore
            if len(mask) > 0 and len(mask) != len(self):  # type: ignore
                msg = f"Invalid mask size {len(mask)}. (expected {len(self)})"
                raise ValueError(msg)
            indices = _where_1d(mask)
            return self.get_items_indices(indices, *args)
        else:
            return self.get_item(mask, *args)

    def get_items_slice(
        self,
        slice_: slice,
        *args,
    ) -> List[T]:
        if self._add_slice_support:
            return self.get_items_indices(range(len(self))[slice_], *args)
        else:
            return self.get_item(slice_, *args)

    def get_items_none(
        self,
        none: None,
        *args,
    ) -> List[T]:
        if self._add_none_support:
            return self.get_items_indices(slice(None), *args)
        else:
            return self.get_item(none, *args)


class DatasetSlicerWrapper(Generic[T], DatasetSlicer[T], Wrapper[T]):
    def __init__(
        self,
        dataset: SupportsLenAndGetItem[T],
        *,
        add_slice_support: bool = True,
        add_indices_support: bool = True,
        add_mask_support: bool = True,
    ) -> None:
        """Wrap a sequence to support slice, indices and mask arguments types."""
        DatasetSlicer.__init__(
            self,
            add_slice_support=add_slice_support,
            add_indices_support=add_indices_support,
            add_mask_support=add_mask_support,
        )
        Wrapper.__init__(self, dataset)

    def __len__(self) -> int:
        return len(self.dataset)

    def get_item(self, idx: int, *args) -> T:
        # note: we need to split calls here, because self.dataset[idx] give an int as argument while self.dataset[idx, *args] always gives a tuple even if args == ()
        if len(args) == 0:
            return self.dataset[idx]
        else:
            # equivalent to self.dataset[idx, *args], but only in recent python versions
            return self.dataset.__getitem__((idx,) + args)


def _where_1d(mask: Union[Iterable[bool], Tensor]) -> Tensor1D:
    if not isinstance(mask, BoolTensor):
        mask = torch.as_tensor(list(mask), dtype=torch.bool)  # type: ignore
    return torch.where(mask)[0]  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import TestCase

from torchoutil.entries import print_safe_rmdir, print_tree


class TestEntries(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

        root = tmpdir.joinpath("test_print_safe_rmdir")
        root.joinpath("a", "a1").mkdir(parents=True)
        root.joinpath("a", "a2").mkdir()
        root.joinpath("b").mkdir()
        root.joinpath("c").mkdir()

        cls.safe_rmdir_root = root

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.safe_rmdir_root.is_dir():
            shutil.rmtree(cls.safe_rmdir_root)

    def test_print_safe_rmdir(self) -> None:
        print_safe_rmdir(self.safe_rmdir_root)

    def test_print_tree(self) -> None:
        print_tree(".", max_depth=0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.core.get import get_device
from torchoutil.nn.functional.multilabel import (
    indices_to_multihot,
    indices_to_multinames,
    multihot_to_indices,
    multihot_to_multinames,
    multinames_to_indices,
    multinames_to_multihot,
    probs_to_indices,
    probs_to_multihot,
    probs_to_multinames,
)


class TestMultilabel(TestCase):
    def test_indices_to_multihot_1(self) -> None:
        indices = [[1, 2], [0], [], [3]]
        num_classes = 4
        expected_multihot = torch.as_tensor(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1]], dtype=torch.bool
        )

        result = indices_to_multihot(indices, num_classes=num_classes)

        assert torch.equal(result, expected_multihot)

    def test_probs_to_indices_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_indices = [[0, 3]]

        result = probs_to_indices(probs, threshold=0.5)

        assert result == expected_indices

    def test_probs_to_multihot_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_multihot = torch.as_tensor([[True, False, False, True]])

        result = probs_to_multihot(probs, threshold=0.5)

        assert torch.equal(result, expected_multihot)

    def test_probs_to_multinames_1(self) -> None:
        probs = torch.as_tensor([[1.0, 0.4, 0.1, 0.9]])
        expected_multinames = [["0", "3"]]

        num_classes = probs.shape[-1]
        idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))
        result = probs_to_multinames(probs, threshold=0.5, idx_to_name=idx_to_name)

        assert result == expected_multinames

    def test_convert_multihot(self) -> None:
        num_samples = int(torch.randint(1, 20, ()).item())
        num_classes = int(torch.randint(1, 20, ()).item())
        threshold = torch.rand(())
        idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))

        probs_1 = torch.rand(num_samples, num_classes)
        multihot_1 = probs_1.ge(threshold)

        indices_1 = multihot_to_indices(multihot_1)
        multinames_1 = indices_to_multinames(indices_1, idx_to_name)
        multihot_2 = multinames_to_multihot(multinames_1, idx_to_name)

        multinames_2 = multihot_to_multinames(multihot_2, idx_to_name)
        indices_2 = multinames_to_indices(multinames_2, idx_to_name)

        assert torch.equal(multihot_1, multihot_2), f"{multihot_1=} ; {multihot_2=}"
        assert multinames_1 == multinames_2
        assert indices_1 == indices_2

    def test_ints_to_multihots(self) -> None:
        device = get_device()
        ints = torch.as_tensor([[0, 1, 1]], device=device)
        num_classes = 5
        multihots = indices_to_multihot(ints, num_classes, dtype=torch.int)
        expected = torch.as_tensor([[1, 1, 0, 0, 0]], device=device)

        assert multihots.shape == expected.shape
        assert multihots.eq(expected).all(), f"{multihots=}"

    def test_convert_and_reconvert(self) -> None:
        device = get_device()
        multihots = torch.as_tensor(
            [
                [1, 1, 1],
                [0, 0, 0],
                [1, 1, 0],
                [0, 0, 1],
            ],
            device=device,
        )
        expected_ints = [[0, 1, 2], [], [0, 1], [2]]

        num_classes = multihots.shape[1]
        ints = multihot_to_indices(multihots)
        new_multihots = indices_to_multihot(
            ints, num_classes, dtype=torch.int, device=device
        )

        assert ints == expected_ints
        assert multihots.shape == new_multihots.shape
        assert multihots.eq(new_multihots).all()

    def test_empty_case_1(self) -> None:
        num_samples = 0
        num_classes = 5

        multihot = torch.rand(num_samples, num_classes).ge(0.5)
        with self.assertRaises(ValueError):
            multihot_to_indices(multihot)

        indices = torch.empty(0, 5)
        with self.assertRaises(ValueError):
            indices_to_multihot(indices, num_classes)

    def test_empty_case_2(self) -> None:
        num_samples = 5
        num_classes = 0
        expected = [[], [], [], [], []]

        multihot = torch.rand(num_samples, num_classes).ge(0.5)

        result_1 = multihot_to_indices(multihot)
        assert result_1 == expected

        result_2 = indices_to_multihot(expected, num_classes)
        assert torch.equal(result_2, multihot)

    def test_empty_case_3(self) -> None:
        num_steps = 3
        num_samples = 0
        num_classes = 5

        multihot = torch.rand(num_steps, num_samples, num_classes).ge(0.5)

        with self.assertRaises(ValueError):
            multihot_to_indices(multihot)

        indices = [torch.empty(0, 5) for _ in range(num_steps)]
        with self.assertRaises(ValueError):
            indices_to_multihot(indices, num_classes)

    def test_probs_to_multihot_dim(self) -> None:
        probs = torch.rand(16, 10, 5)
        threshold = torch.full((10,), 0.3)

        multihot = probs_to_multihot(probs, threshold, dim=1)
        assert multihot.shape == probs.shape

    def test_probs_to_indices_dim(self) -> None:
        # 2x5
        probs = torch.as_tensor(
            [
                [0.6, 0.9, 0.1, 0.0, 1.0],
                [0.0, 0.1, 0.9, 0.0, 1.0],
            ]
        )
        indices_0 = probs_to_indices(probs, 0.5, dim=0)
        assert indices_0 == [[0], [0], [1], [], [0, 1]]

        indices_1 = probs_to_indices(probs, 0.5, dim=1)
        assert indices_1 == [[0, 1, 4], [2, 4]]

        # duplicate probs for 3D test with 2x2x5
        probs = torch.stack([probs, probs])

        indices_0 = probs_to_indices(probs, 0.5, dim=0)
        assert indices_0 == [
            [[0, 1], []],
            [[0, 1], []],
            [[], [0, 1]],
            [[], []],
            [[0, 1], [0, 1]],
        ]

        indices_1 = probs_to_indices(probs, 0.5, dim=1)
        assert indices_1 == [[[0], [0], [1], [], [0, 1]], [[0], [0], [1], [], [0, 1]]]

        indices_2 = probs_to_indices(probs, 0.5, dim=2)
        assert indices_2 == [[[0, 1, 4], [2, 4]], [[0, 1, 4], [2, 4]]]


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch.nn.modules import *

from .activation import LogSoftmaxMultidim, SoftmaxMultidim
from .crop import CropDim, CropDims
from .layer import PositionalEncoding
from .mask import MaskedMean, MaskedSum
from .mixins import EModule, EModuleDict, EModuleList, EModulePartial, ESequential
from .multiclass import (
    IndexToName,
    IndexToOnehot,
    NameToIndex,
    NameToOnehot,
    OnehotToIndex,
    OnehotToName,
    ProbsToIndex,
    ProbsToName,
    ProbsToOnehot,
)
from .multilabel import (
    IndicesToMultihot,
    IndicesToMultinames,
    MultihotToIndices,
    MultihotToMultinames,
    MultinamesToIndices,
    MultinamesToMultihot,
    ProbsToIndices,
    ProbsToMultihot,
    ProbsToMultinames,
)
from .numpy import NumpyToTensor, TensorToNumpy, ToNumpy
from .pad import PadAndStackRec, PadDim, PadDims
from .powerset import MultilabelToPowerset, PowersetToMultilabel
from .tensor import (
    FFT,
    IFFT,
    Abs,
    Angle,
    AsTensor,
    Exp,
    Exp2,
    Imag,
    Log,
    Log2,
    Log10,
    Max,
    Mean,
    Min,
    Normalize,
    Permute,
    Pow,
    Real,
    Repeat,
    RepeatInterleave,
    Reshape,
    Squeeze,
    TensorTo,
    ToItem,
    ToList,
    Topk,
    Transpose,
    Unsqueeze,
    View,
)
from .transform import (
    Flatten,
    Identity,
    PadAndCropDim,
    RepeatInterleaveNd,
    ResampleNearestFreqs,
    ResampleNearestRates,
    ResampleNearestSteps,
    Shuffled,
    ToTensor,
    TransformDrop,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import Callable, Generic, Iterable, Iterator, Optional, TypeVar, Union

from torch.utils.data.dataset import Dataset, IterableDataset
from torch.utils.data.dataset import Subset as TorchSubset

from torchoutil import LongTensor1D
from torchoutil.pyoutil.collections import is_sorted
from torchoutil.pyoutil.typing.classes import (
    SupportsLenAndGetItem,
    SupportsLenAndGetItemAndIter,
)

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)

SizedDatasetLike = SupportsLenAndGetItem
SizedIterableDatasetLike = SupportsLenAndGetItemAndIter

T_Dataset = TypeVar("T_Dataset", bound=Dataset)
T_SizedDatasetLike = TypeVar("T_SizedDatasetLike", bound=SupportsLenAndGetItem)
T_SizedIterableDataset = TypeVar(
    "T_SizedIterableDataset",
    bound=SupportsLenAndGetItemAndIter,
)


class EmptyDataset(Dataset[None]):
    """Dataset placeholder. Raises StopIteration if __getitem__ is called."""

    def __getitem__(self, idx, /) -> None:
        raise StopIteration

    def __len__(self) -> int:
        return 0


class Wrapper(Generic[T], Dataset[T]):
    def __init__(self, dataset: SupportsLenAndGetItem[T]) -> None:
        Dataset.__init__(self)
        self.dataset = dataset

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, idx, /) -> T:
        raise NotImplementedError

    def unwrap(self, recursive: bool = True) -> Union[SupportsLenAndGetItem, Dataset]:
        dataset = self.dataset
        continue_ = recursive and isinstance(dataset, Wrapper)
        while continue_:
            if not isinstance(dataset, (Wrapper, TorchSubset)):
                break
            dataset = dataset.dataset
            continue_ = isinstance(dataset, Wrapper)
        return dataset

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.dataset)})"


class IterableWrapper(Generic[T], IterableDataset[T], Wrapper[T]):
    def __init__(self, dataset: SupportsLenAndGetItem[T]) -> None:
        IterableDataset.__init__(self)
        Wrapper.__init__(self, dataset)

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        ...

    def _get_dataset_iter(self) -> Iterator[T]:
        if hasattr(self.dataset, "__iter__"):
            it = iter(self.dataset)
        else:
            it = (self.dataset[i] for i in range(len(self.dataset)))
        return it


class TransformWrapper(Generic[T, U], Wrapper[T]):
    def __init__(
        self,
        dataset: SupportsLenAndGetItem[T],
        transform: Optional[Callable[[T], U]],
        condition: Optional[Callable[[T, int], bool]] = None,
    ) -> None:
        super().__init__(dataset)
        self._transform = transform
        self._condition = condition

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx) -> Union[T, U]:
        assert isinstance(idx, int)
        item = self.dataset[idx]
        if self._transform is not None and (
            self._condition is None or self._condition(item, idx)
        ):
            item = self._transform(item)
        return item

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    @property
    def condition(self) -> Optional[Callable[[T, int], bool]]:
        return self._condition


class IterableTransformWrapper(IterableWrapper[T], Generic[T, U]):
    def __init__(
        self,
        dataset: SupportsLenAndGetItem[T],
        transform: Optional[Callable[[T], U]],
        condition: Optional[Callable[[T, int], bool]] = None,
    ) -> None:
        super().__init__(dataset)
        self._transform = transform
        self._condition = condition

    def __len__(self) -> int:
        return len(self.dataset)

    def __iter__(self) -> Iterator[Union[T, U]]:
        it = super()._get_dataset_iter()
        for i, item in enumerate(it):
            if self._transform is not None and (
                self._condition is None or self._condition(item, i)
            ):
                item = self._transform(item)
            yield item
        return

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    @property
    def condition(self) -> Optional[Callable[[T, int], bool]]:
        return self._condition


class IterableSubset(IterableWrapper[T], Generic[T]):
    def __init__(
        self,
        dataset: SupportsLenAndGetItem[T],
        indices: Union[Iterable[int], LongTensor1D],
    ) -> None:
        if isinstance(indices, LongTensor1D):
            indices = indices.tolist()
        else:
            indices = list(indices)

        if not all(idx >= 0 for idx in indices) or not is_sorted(indices):
            msg = f"Invalid argument {indices=}. (expected a sorted list of positive integers)"
            raise ValueError(msg)

        super().__init__(dataset)
        self._indices = indices

    def __len__(self) -> int:
        return len(self._indices)

    def __iter__(self) -> Iterator[T]:
        it = super()._get_dataset_iter()

        cur_idx = 0
        item = next(it)

        for idx in self._indices:
            if cur_idx == idx:
                yield item
                continue

            while cur_idx < idx:
                cur_idx += 1
                item = next(it)

            yield item
        return


class Subset(Generic[T], TorchSubset[T], Wrapper[T]):
    def __init__(self, dataset: SizedDatasetLike[T], indices: Iterable[int]) -> None:
        indices = list(indices)
        TorchSubset.__init__(self, dataset, indices)  # type: ignore
        Wrapper.__init__(self, dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle
from pathlib import Path
from typing import Any


def load_pickle(fpath: Path) -> Any:
    content = pickle.loads(fpath.read_bytes())
    return content


def to_pickle(obj: Any, fpath: Path) -> bytes:
    content = pickle.dumps(obj)
    fpath.write_bytes(content)
    return content



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os

from torchoutil.core.packaging import _COLORLOG_AVAILABLE

if not _COLORLOG_AVAILABLE:
    msg = "Cannot import colorlog objects because optional dependancy 'colorlog' is not installed. Please install it using 'pip install torchoutil[extras]'"
    raise ImportError(msg)

from colorlog import ColoredFormatter  # type: ignore

pylog = logging.getLogger(__name__)


LOG_COLORS = {
    "DEBUG": "purple",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "bold_red",
}


def get_colored_formatter() -> ColoredFormatter:
    if not _COLORLOG_AVAILABLE:
        msg = "Cannot call function get_colored_formatter because optional dependancy 'colorlog' is not installed. Please install it using 'pip install torchoutil[extras]'"
        raise RuntimeError(msg)

    rank = os.getenv("SLURM_PROCID", 0)
    fmt = f"[%(purple)sRANK{rank}%(reset)s][%(cyan)s%(asctime)s%(reset)s][%(blue)s%(name)s%(reset)s][%(log_color)s%(levelname)s%(reset)s] - %(message)s"
    formatter = ColoredFormatter(fmt=fmt, log_colors=LOG_COLORS)
    return formatter



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .abc import Singleton
from .argparse import (
    str_to_bool,
    str_to_optional_bool,
    str_to_optional_int,
    str_to_optional_str,
)
from .collections import (
    KeyMode,
    all_eq,
    all_ne,
    argmax,
    argmin,
    contained,
    dict_list_to_list_dict,
    dump_dict,
    filter_iterable,
    flat_dict_of_dict,
    flat_list_of_list,
    flatten,
    intersect_lists,
    list_dict_to_dict_list,
    prod,
    recursive_generator,
    sorted_dict,
    unflat_dict_of_dict,
    unflat_list_of_list,
    union_dicts,
    union_lists,
    unzip,
)
from .csv import load_csv, to_csv
from .dataclasses import get_defaults_values
from .datetime import now_iso
from .difflib import find_closest_in_list, sequence_matcher_ratio
from .enum import StrEnum
from .functools import Compose, compose, identity
from .hashlib import hash_file
from .importlib import (
    package_is_available,
    reload_globals_modules,
    reload_submodules,
    search_imported_submodules,
)
from .inspect import get_current_fn_name, get_fullname
from .io import open_close_wrap
from .json import load_json, to_json
from .logging import (
    VERBOSE_DEBUG,
    VERBOSE_ERROR,
    VERBOSE_INFO,
    VERBOSE_WARNING,
    MkdirFileHandler,
    get_current_file_logger,
    get_ipython_name,
    get_null_logger,
    running_on_interpreter,
    running_on_notebook,
    running_on_terminal,
    setup_logging_level,
    setup_logging_verbose,
    warn_once,
)
from .math import clamp, clip
from .os import get_num_cpus_available, safe_rmdir, tree_iter
from .re import (
    PatternLike,
    PatternListLike,
    compile_patterns,
    find_patterns,
    get_key_fn,
    match_patterns,
    sort_with_patterns,
)
from .typing import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import Namespace
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Iterable, Literal, Mapping, Optional, Type, Union

from yaml import (
    BaseLoader,
    CBaseLoader,
    CFullLoader,
    CLoader,
    CSafeLoader,
    CUnsafeLoader,
    FullLoader,
    Loader,
    MappingNode,
    Node,
    SafeLoader,
    ScalarNode,
    SequenceNode,
    UnsafeLoader,
)
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from torchoutil.core.packaging import _OMEGACONF_AVAILABLE, _YAML_AVAILABLE
from torchoutil.pyoutil.typing import DataclassInstance, NamedTupleInstance
from torchoutil.utils.saving.common import to_builtin

if not _YAML_AVAILABLE:
    msg = f"Cannot use python module {__file__} since pyyaml package is not installed."
    raise ImportError(msg)

import yaml

if _OMEGACONF_AVAILABLE:
    from omegaconf import OmegaConf  # type: ignore


YamlLoaders = Union[
    Type[Loader],
    Type[BaseLoader],
    Type[FullLoader],
    Type[SafeLoader],
    Type[UnsafeLoader],
    Type[CLoader],
    Type[CBaseLoader],
    Type[CFullLoader],
    Type[CSafeLoader],
    Type[CUnsafeLoader],
]


def to_yaml(
    data: Union[
        Iterable[Any],
        Mapping[str, Any],
        Namespace,
        DataclassInstance,
        NamedTupleInstance,
    ],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    resolve: bool = False,
    encoding: Optional[str] = "utf-8",
    # YAML dump kwargs
    sort_keys: bool = False,
    indent: Union[int, None] = None,
    width: Union[int, None] = 1000,
    allow_unicode: bool = True,
    **yaml_dump_kwds,
) -> str:
    """Dump content to yaml format."""
    if not _OMEGACONF_AVAILABLE and resolve:
        raise ValueError(
            "Cannot resolve yaml config without omegaconf package."
            "Please use resolve=False or install omegaconf with 'pip install torchoutil[extras]'."
        )

    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    if resolve:
        data = OmegaConf.create(data)  # type: ignore
        data = OmegaConf.to_container(data, resolve=True)  # type: ignore

    if to_builtins:
        data = to_builtin(data)

    content = yaml.dump(
        data,
        sort_keys=sort_keys,
        indent=indent,
        width=width,
        allow_unicode=allow_unicode,
        **yaml_dump_kwds,
    )
    if fpath is not None:
        fpath.write_text(content, encoding=encoding)
    return content


def load_yaml(
    fpath: Union[str, Path, TextIOWrapper],
    *,
    Loader: YamlLoaders = SafeLoader,
    on_error: Literal["raise", "ignore"] = "raise",
) -> Any:
    """Load content from yaml filepath."""
    if isinstance(fpath, (str, Path)):
        with open(fpath, "r") as file:
            return load_yaml(file, Loader=Loader, on_error=on_error)

    try:
        data = yaml.load(fpath, Loader=Loader)
    except (ScannerError, ParserError) as err:
        if on_error == "ignore":
            return None
        else:
            raise err
    return data


class IgnoreTagLoader(SafeLoader):
    """SafeLoader that ignores yaml tags.

    Examples
    ========

    ```python
    >>> dumped = "a: !!python/tuple\n- 1\n- 2"
    >>> yaml.load(dumped, Loader=IgnoreTagLoader)
    ... {"a": [1, 2]}
    >>> yaml.load(dumped, Loader=FullLoader)
    ... {"a": (1, 2)}
    >>> yaml.load(dumped, Loader=SafeLoader)  # raises ConstructorError
    ```
    """

    def construct_with_tag(self, tag: str, node: Node) -> Any:
        if isinstance(node, MappingNode):
            return self.construct_mapping(node)
        elif isinstance(node, ScalarNode):
            return self.construct_scalar(node)
        elif isinstance(node, SequenceNode):
            return self.construct_sequence(node)
        else:
            raise NotImplementedError(f"Unsupported node type {type(node)}.")


IgnoreTagLoader.add_multi_constructor("!", IgnoreTagLoader.construct_with_tag)
IgnoreTagLoader.add_multi_constructor("tag:", IgnoreTagLoader.construct_with_tag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from typing import TypeVar

T = TypeVar("T")


def clip(x: T, xmin: T = -math.inf, xmax: T = math.inf) -> T:
    return min(max(x, xmin), xmax)


def clamp(x: T, xmin: T = -math.inf, xmax: T = math.inf) -> T:
    return clip(x, xmin, xmax)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil import (
    get_inverse_perm,
    insert_at_indices,
    lengths_to_non_pad_mask,
    masked_mean,
    multihot_to_indices,
    probs_to_multinames,
    probs_to_name,
)


class TestReadme(TestCase):
    def test_masked_mean_example(self) -> None:
        x = torch.as_tensor([1, 2, 3, 4])
        mask = torch.as_tensor([True, True, False, False])
        result = masked_mean(x, mask)
        # result contains the mean of the values marked as True: 1.5
        result = result.item()

        assert result == 1.5

    def test_lengths_to_non_pad_mask_example(self) -> None:
        x = torch.as_tensor([3, 1, 2])
        pad_mask = lengths_to_non_pad_mask(x, max_len=4)
        expected = torch.as_tensor(
            [
                [True, True, True, False],
                [True, False, False, False],
                [True, True, False, False],
            ]
        )

        assert torch.equal(pad_mask, expected)

    def test_probs_to_name_example(self) -> None:
        probs = torch.as_tensor([[0.9, 0.1], [0.4, 0.6]])
        names = probs_to_name(probs, idx_to_name={0: "Cat", 1: "Dog"})
        expected = ["Cat", "Dog"]

        assert names == expected

    def test_probs_to_multinames_example(self) -> None:
        probs = torch.as_tensor([[0.9, 0.1], [0.6, 0.9]])
        names = probs_to_multinames(
            probs, threshold=0.5, idx_to_name={0: "Cat", 1: "Dog"}
        )
        expected = [["Cat"], ["Cat", "Dog"]]

        assert names == expected

    def test_multihot_to_indices_example(self) -> None:
        multihot = torch.as_tensor([[1, 0, 0], [0, 1, 1], [0, 0, 0]])
        indices = multihot_to_indices(multihot)
        expected = [[0], [1, 2], []]

        assert indices == expected

    def test_insert_at_indices_example(self) -> None:
        x = torch.as_tensor([1, 2, 3, 4])
        result = insert_at_indices(x, [0, 2], 5)
        expected = torch.as_tensor([5, 1, 2, 5, 3, 4])

        assert torch.equal(result, expected)

    def test_get_inverse_perm_example(self) -> None:
        perm = torch.randperm(10)
        inv_perm = get_inverse_perm(perm)

        x1 = torch.rand(10)
        x2 = x1[perm]
        x3 = x2[inv_perm]
        # inv_perm are indices that allow us to get x1 from x3, i.e. x1 == x3 here

        assert torch.equal(x1, x3)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from numbers import Integral, Number
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from typing_extensions import TypeGuard, TypeIs

from .classes import (
    BuiltinNumber,
    BuiltinScalar,
    DataclassInstance,
    NamedTupleInstance,
    NoneType,
)


def is_builtin_obj(x: Any) -> bool:
    """Returns True if object is an instance of a builtin object.

    Note: If the object is an instance of a custom subtype of a builtin object, this function returns False.
    """
    return x.__class__.__module__ == "builtins" and not isinstance(x, type)


def is_builtin_number(x: Any, *, strict: bool = False) -> TypeIs[BuiltinNumber]:
    """Returns True if x is an instance of a builtin number type (int, float, bool, complex).

    Args:
        x: Object to check.
        strict: If True, it will not consider subtypes of builtins as builtin numbers.
    """
    if strict and not is_builtin_obj(x):
        return False
    return isinstance(x, (int, float, bool, complex))


def is_builtin_scalar(x: Any, *, strict: bool = False) -> TypeIs[BuiltinScalar]:
    """Returns True if x is an instance of a builtin scalar type (int, float, bool, complex, NoneType, str, bytes).

    Args:
        x: Object to check.
        strict: If True, it will not consider subtypes of builtins as builtin numbers.
    """
    if strict and not is_builtin_obj(x):
        return False
    return isinstance(x, (int, float, bool, complex, NoneType, str, bytes))


def is_dataclass_instance(x: Any) -> TypeIs[DataclassInstance]:
    """Returns True if argument is a dataclass. Unlike function `dataclasses.is_dataclass`, this function returns False for a dataclass type."""
    return not isinstance(x, type) and isinstance(x, DataclassInstance)


def is_dict_str(x: Any) -> TypeIs[Dict[str, Any]]:
    return isinstance(x, dict) and all(isinstance(key, str) for key in x.keys())


def is_dict_str_optional_int(x: Any) -> TypeIs[Dict[str, Optional[int]]]:
    return (
        isinstance(x, dict)
        and all(isinstance(key, str) for key in x.keys())
        and all(isinstance(value, (int, NoneType)) for value in x.values())
    )


def is_iterable_bool(
    x: Any,
    *,
    accept_generator: bool = True,
) -> TypeIs[Iterable[bool]]:
    if not accept_generator and isinstance(x, Generator):
        return False
    return isinstance(x, Iterable) and all(isinstance(xi, bool) for xi in x)


def is_iterable_bytes_or_list(
    x: Any,
    *,
    accept_generator: bool = True,
) -> TypeIs[Iterable[Union[bytes, list]]]:
    if not accept_generator and isinstance(x, Generator):
        return False
    return isinstance(x, Iterable) and all(isinstance(xi, (bytes, list)) for xi in x)


def is_iterable_float(
    x: Any,
    *,
    accept_generator: bool = True,
) -> TypeIs[Iterable[float]]:
    if not accept_generator and isinstance(x, Generator):
        return False
    return isinstance(x, Iterable) and all(isinstance(xi, float) for xi in x)


def is_iterable_int(
    x: Any,
    *,
    accept_bool: bool = True,
    accept_generator: bool = True,
) -> TypeIs[Iterable[int]]:
    if not accept_generator and isinstance(x, Generator):
        return False
    return isinstance(x, Iterable) and all(
        isinstance(xi, int) and (accept_bool or not isinstance(xi, bool)) for xi in x
    )


def is_iterable_integral(
    x: Any,
    *,
    accept_generator: bool = True,
) -> TypeIs[Iterable[Integral]]:
    if not accept_generator and isinstance(x, Generator):
        return False
    return isinstance(x, Iterable) and all(isinstance(xi, Integral) for xi in x)


def is_iterable_iterable_int(x: Any) -> TypeIs[Iterable[Iterable[int]]]:
    return (
        isinstance(x, Iterable)
        and all(isinstance(xi, Iterable) for xi in x)
        and all(isinstance(xij, int) for xi in x for xij in xi)
    )


def is_iterable_str(
    x: Any,
    *,
    accept_str: bool = True,
    accept_generator: bool = True,
) -> TypeGuard[Iterable[str]]:
    if isinstance(x, str):
        return accept_str
    if isinstance(x, Generator):
        return accept_generator and all(isinstance(xi, str) for xi in x)
    return isinstance(x, Iterable) and all(isinstance(xi, str) for xi in x)


def is_list_list_str(x: Any) -> TypeIs[List[List[str]]]:
    return (
        isinstance(x, list)
        and all(isinstance(xi, list) for xi in x)
        and all(isinstance(xij, str) for xi in x for xij in xi)
    )


def is_list_bool(x: Any) -> TypeIs[List[bool]]:
    return isinstance(x, list) and all(isinstance(xi, bool) for xi in x)


def is_list_float(x: Any) -> TypeIs[List[float]]:
    return isinstance(x, list) and (all(isinstance(xi, float) for xi in x))


def is_list_int(x: Any) -> TypeIs[List[int]]:
    return isinstance(x, list) and all(isinstance(xi, int) for xi in x)


def is_list_number(x: Any) -> TypeIs[List[Number]]:
    return isinstance(x, list) and all(isinstance(xi, Number) for xi in x)


def is_list_builtin_number(x: Any) -> TypeIs[List[BuiltinNumber]]:
    return isinstance(x, list) and all(is_builtin_number(xi) for xi in x)


def is_list_str(x: Any) -> TypeIs[List[str]]:
    return isinstance(x, list) and all(isinstance(xi, str) for xi in x)


def is_mapping_str(x: Any) -> TypeIs[Mapping[str, Any]]:
    return isinstance(x, Mapping) and all(isinstance(key, str) for key in x.keys())


def is_namedtuple_instance(x: Any) -> TypeIs[NamedTupleInstance]:
    return not isinstance(x, type) and isinstance(x, NamedTupleInstance)


def is_sequence_bool(x: Any) -> TypeIs[Sequence[bool]]:
    return isinstance(x, Sequence) and all(isinstance(xi, bool) for xi in x)


def is_sequence_int(x: Any) -> TypeIs[Sequence[int]]:
    return isinstance(x, Sequence) and all(isinstance(xi, int) for xi in x)


def is_sequence_str(
    x: Any,
    *,
    accept_str: bool = True,
) -> TypeGuard[Sequence[str]]:
    return (accept_str and isinstance(x, str)) or (
        not isinstance(x, str)
        and isinstance(x, Sequence)
        and all(isinstance(xi, str) for xi in x)
    )


def is_tuple_optional_int(x: Any) -> TypeIs[Tuple[Optional[int], ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, (int, NoneType)) for xi in x)


def is_tuple_int(x: Any) -> TypeIs[Tuple[int, ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, int) for xi in x)


def is_tuple_str(x: Any) -> TypeIs[Tuple[str, ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, str) for xi in x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import unittest
from unittest import TestCase

import torch

from torchoutil import pyoutil as po
from torchoutil.nn.functional.checksum import checksum


class TestChecksum(TestCase):
    def test_checksum_tensor(self) -> None:
        x = [
            torch.arange(10),
            torch.arange(10).view(2, 5),
            [torch.arange(10)],
            list(range(10)),
            tuple(range(10)),
            set(range(10)),
            range(10),
            [],
            None,
            0,
            1,
            math.nan,
            [1, 2],
            [2, 1],
            [1, 2, 0],
            (1, 2),
        ]
        csums = [checksum(xi) for xi in x]
        assert po.all_ne(csums), f"{csums=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import hashlib
import logging
from pathlib import Path
from typing import Literal, Protocol, Union, runtime_checkable

from typing_extensions import Buffer

HashName = Literal["sha256", "md5"]

DEFAULT_CHUNK_SIZE = 256 * 1024**2  # 256 MiB
HASH_NAMES = ("sha256", "md5")

pylog = logging.getLogger(__name__)


@runtime_checkable
class Hasher(Protocol):
    digest_size: int
    block_size: int
    name: str

    def hexdigest(self) -> str:
        ...

    def update(self, data: Buffer, /) -> None:
        ...


def hash_file(
    fpath: Union[str, Path],
    hash_type: Union[HashName, Hasher],
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> str:
    """Return the hash value for a file.

    BASED ON https://github.com/pytorch/audio/blob/v0.13.0/torchaudio/datasets/utils.py#L110
    """
    if hash_type == "sha256":
        hasher = hashlib.sha256()
    elif hash_type == "md5":
        hasher = hashlib.md5()
    elif isinstance(hash_type, Hasher):
        hasher = hash_type
    else:
        msg = f"Invalid argument hash_type={hash_type}. (expected one of {HASH_NAMES} or a custom hasher)"
        raise ValueError(msg)
    del hash_type

    with open(fpath, "rb") as file:
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            hasher.update(chunk)

    hash_value = hasher.hexdigest()
    return hash_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Iterable, Union

from torchoutil.pyoutil.argparse import str_to_bool
from torchoutil.pyoutil.os import safe_rmdir, tree_iter
from torchoutil.pyoutil.re import PatternLike

pylog = logging.getLogger(__name__)


def print_tree(
    root: Union[str, Path],
    *,
    exclude: Union[PatternLike, Iterable[PatternLike]] = (),
    max_depth: int = sys.maxsize,
    followlinks: bool = False,
) -> None:
    num_dirs = 0
    num_files = 0
    for line in tree_iter(
        root=root,
        exclude=exclude,
        max_depth=max_depth,
        followlinks=followlinks,
    ):
        print(f"{line}")

        if line.endswith("/"):
            num_dirs += 1
        else:
            num_files += 1

    print(f"\n{num_dirs} directories, {num_files} files")


def main_tree() -> None:
    parser = ArgumentParser()
    parser.add_argument(
        "root",
        type=str,
        help="Root directory path.",
        default=".",
        nargs="?",  # for optional positional argument
    )
    parser.add_argument(
        "--exclude",
        type=str,
        help="Exclude file patterns.",
        default=(),
        nargs="*",
    )
    parser.add_argument(
        "--max_depth",
        type=int,
        help="Max directory tree depth.",
        default=sys.maxsize,
    )
    parser.add_argument(
        "--followlinks",
        type=str_to_bool,
        help="Indicates whether or not symbolic links shound be followed. defaults to True.",
        default=True,
    )
    args = parser.parse_args()

    print_tree(
        root=args.root,
        exclude=args.exclude,
        max_depth=args.max_depth,
        followlinks=args.followlinks,
    )


def print_safe_rmdir(
    root: Union[str, Path],
    *,
    rm_root: bool = True,
    error_on_non_empty_dir: bool = True,
    followlinks: bool = False,
    dry_run: bool = False,
    verbose: int = 0,
) -> None:
    deleted, reviewed = safe_rmdir(
        root=root,
        rm_root=rm_root,
        error_on_non_empty_dir=error_on_non_empty_dir,
        followlinks=followlinks,
        dry_run=dry_run,
        verbose=verbose,
    )
    if dry_run:
        print(
            f"Dry run mode enabled. Here is the list of directories to delete ({len(deleted)}/{len(reviewed)}):"
        )
        for path in deleted:
            print(f" - {path}")

    elif verbose >= 1:
        print(f"{len(deleted)} directories has been deleted.")


def main_safe_rmdir() -> None:
    parser = ArgumentParser()
    parser.add_argument(
        "root",
        type=str,
        help="Root directory path.",
    )
    parser.add_argument(
        "--rm_root",
        type=str_to_bool,
        default=True,
        help="If True, remove the root directory too if it is empty at the end. defaults to True.",
    )
    parser.add_argument(
        "--error_on_non_empty_dir",
        type=str_to_bool,
        default=True,
        help="If True, raises a RuntimeError if a subdirectory contains at least 1 file. Otherwise it will ignore non-empty directories. defaults to True.",
    )
    parser.add_argument(
        "--followlinks",
        type=str_to_bool,
        default=False,
        help="Indicates whether or not symbolic links shound be followed. defaults to False.",
    )
    parser.add_argument(
        "--dry_run",
        type=str_to_bool,
        default=False,
        help="If True, does not remove any directory and just output the list of directories which could be deleted. defaults to False.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=0,
        help="Verbose level. defaults to 0.",
    )
    args = parser.parse_args()
    print_safe_rmdir(
        root=args.root,
        rm_root=args.rm_root,
        error_on_non_empty_dir=args.error_on_non_empty_dir,
        followlinks=args.followlinks,
        dry_run=args.dry_run,
        verbose=args.verbose,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import sys
from dataclasses import asdict, dataclass
from typing import Any, Dict, Tuple, TypedDict, Union, overload

from typing_extensions import NotRequired, TypeIs

from torchoutil.pyoutil.typing import NoneType, is_dict_str_optional_int

# Pattern of https://semver.org/
_VERSION_PATTERN = r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
_VERSION_FORMAT = r"{major}.{minor}.{patch}"
_VERSION_KEYS = ("major", "minor", "patch", "prerelease", "buildmetadata")


PreRelease = Union[int, str, None, list]
BuildMetadata = Union[int, str, None, list]


class VersionDict(TypedDict):
    major: int
    minor: int
    patch: int
    prerelease: NotRequired[PreRelease]
    buildmetadata: NotRequired[BuildMetadata]


VersionTupleLike = Union[
    Tuple[int, int, int],
    Tuple[int, int, int, PreRelease],
    Tuple[int, int, int, PreRelease, BuildMetadata],
]


@dataclass(init=False, eq=False)
class Version:
    """Version utility class following Semantic Versioning (SemVer) spec.

    Version format is: MAJOR.MINOR.PATCH[-PRERELEASE][+BUILDMETADATA]
    """

    major: int
    minor: int
    patch: int
    prerelease: Union[int, str, None, list]
    buildmetadata: Union[int, str, None, list]

    @overload
    def __init__(
        self,
        version_str: str,
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        version_dict: Dict[str, Union[int, str, None, list]],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        version_tuple: VersionTupleLike,
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        major: int,
        minor: int,
        patch: int,
        prerelease: Union[int, str, None, list] = None,
        buildmetadata: Union[int, str, None, list] = None,
    ) -> None:
        ...

    def __init__(self, *args, **kwargs) -> None:
        # Version str
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], str):
            version_str = args[0]
            version_dict = _parse_version_str(version_str)

        # Version dict
        elif len(args) == 1 and len(kwargs) == 0 and is_dict_str_optional_int(args[0]):
            version_dict = args[0]

        # Version tuple
        elif len(args) == 1 and len(kwargs) == 0 and _is_version_tuple(args[0]):
            version_tuple = args[0]
            version_dict = dict(zip(_VERSION_KEYS, version_tuple))

        # Version args/kwargs
        else:
            version_dict = dict(zip(_VERSION_KEYS, args))
            intersection = set(version_dict.keys()).intersection(kwargs.keys())
            if len(intersection) > 0:
                msg = f"Duplicated argument(s) {tuple(intersection)}. (with {args=} and {kwargs=})"
                raise ValueError(msg)
            version_dict.update(kwargs)  # type: ignore

        major = version_dict["major"]
        minor = version_dict["minor"]
        patch = version_dict["patch"]
        prerelease = version_dict.get("prerelease", None)
        buildmetadata = version_dict.get("buildmetadata", None)

        self.major = major  # type: ignore
        self.minor = minor  # type: ignore
        self.patch = patch  # type: ignore
        self.prerelease = prerelease  # type: ignore
        self.buildmetadata = buildmetadata  # type: ignore

    @classmethod
    def python(cls) -> "Version":
        return Version(
            sys.version_info.major,
            sys.version_info.minor,
            sys.version_info.micro,
        )

    @classmethod
    def from_dict(cls, version_dict: Dict[str, int]) -> "Version":
        return Version(**version_dict)

    @classmethod
    def from_str(cls, version_str: str) -> "Version":
        version_dict = _parse_version_str(version_str)
        return Version(**version_dict)

    @classmethod
    def from_tuple(
        cls,
        version_tuple: VersionTupleLike,
    ) -> "Version":
        return Version(*version_tuple)

    def to_dict(self, exclude_none: bool = True) -> VersionDict:
        version_dict = asdict(self)
        if exclude_none:
            version_dict = {k: v for k, v in version_dict.items() if v is not None}
        return version_dict  # type: ignore

    def to_str(self) -> str:
        kwds = dict(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
        )
        version_str = _VERSION_FORMAT.format(**kwds)
        if self.prerelease is not None:
            version_str = f"{version_str}-{self.prerelease}"
        if self.buildmetadata is not None:
            version_str = f"{version_str}+{self.buildmetadata}"

        return version_str

    def to_tuple(
        self,
        exclude_none: bool = True,
    ) -> VersionTupleLike:
        version_tuple = tuple(self.to_dict().values())
        if exclude_none:
            version_tuple = tuple(v for v in version_tuple if v is not None)
        return version_tuple  # type: ignore

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, (dict, tuple, str)):
            other = Version(other)
        elif not isinstance(other, Version):
            return False

        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
            and self.prerelease == other.prerelease
            and self.buildmetadata == other.buildmetadata
        )

    def __lt__(self, other: "Version") -> bool:
        self_tuple = self.to_tuple()
        other_tuple = other.to_tuple()

        for self_v, other_v in zip(self_tuple, other_tuple):
            if self_v == other_v:
                continue

            if isinstance(self_v, (int, str, NoneType)):
                self_v = [self_v]
            if isinstance(other_v, (int, str, NoneType)):
                other_v = [other_v]

            minlen = min(len(self_v), len(other_v))
            if len(self_v) != len(other_v) and self_v[:minlen] == other_v[:minlen]:
                return len(self_v) < len(other_v)

            for self_vi, other_vi in zip(self_v, other_v):
                if self_vi == other_vi:
                    continue
                if isinstance(self_vi, int) and isinstance(other_vi, int):
                    return self_vi < other_vi
                if isinstance(self_vi, int) and isinstance(other_vi, str):
                    return True
                if isinstance(self_vi, str) and isinstance(other_vi, int):
                    return False
                if isinstance(self_vi, str) and isinstance(other_vi, str):
                    return self_vi < other_vi

                raise TypeError(f"Invalid attribute type {self_vi=} and {other_vi=}.")

        return True

    def __le__(self, other: "Version") -> bool:
        return self == other or self < other

    def __gt__(self, other: "Version") -> bool:
        return self != other and not (self < other)

    def __ge__(self, other: "Version") -> bool:
        return not (self < other)


def _parse_version_str(version_str: str) -> VersionDict:
    version_match = re.match(_VERSION_PATTERN, version_str)
    if version_match is None:
        msg = f"Invalid argument {version_str=}. (not a version)"
        raise ValueError(msg)

    version_dict = version_match.groupdict()
    result = {}
    for k, v in version_dict.items():
        if isinstance(v, str) and "." in v:
            v = v.split(".")
        else:
            v = [v]

        v = [int(vi) if isinstance(vi, str) and vi.isdigit() else vi for vi in v]
        if len(v) == 1:
            v = v[0]
        result[k] = v
    return result  # type: ignore


def _is_version_tuple(
    x: Any,
) -> TypeIs[VersionTupleLike]:
    if not isinstance(x, tuple):
        return False
    if len(x) not in (1, 2, 3):
        return False

    if not isinstance(x[0], int):
        return False
    elif len(x) == 1:
        return True

    if not isinstance(x[1], int):
        return False
    elif len(x) == 2:
        return True

    return isinstance(x[2], int)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Collection of functions and modules to help development in PyTorch.
"""

__name__ = "torchoutil"
__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__status__ = "Development"
__version__ = "0.5.0"


from torch import *  # type: ignore

# Re-import for language servers
from . import core as core
from . import extras as extras
from . import fft as fft
from . import hub as hub
from . import nn as nn
from . import optim as optim
from . import utils as utils
from .core.dtype_enum import DTypeEnum
from .core.dtype_enum import DTypeEnum as dtype_enum
from .core.get import (  # noqa: F401
    CUDA_IF_AVAILABLE,
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    get_device,
    get_dtype,
    get_generator,
)
from .core.semver import Version
from .nn.functional import *
from .types.tensor_subclasses import (
    BoolTensor,
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    ByteTensor,
    ByteTensor0D,
    ByteTensor1D,
    ByteTensor2D,
    ByteTensor3D,
    CDoubleTensor,
    CDoubleTensor0D,
    CDoubleTensor1D,
    CDoubleTensor2D,
    CDoubleTensor3D,
    CFloatTensor,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    CHalfTensor,
    CHalfTensor0D,
    CHalfTensor1D,
    CHalfTensor2D,
    CHalfTensor3D,
    CharTensor,
    CharTensor0D,
    CharTensor1D,
    CharTensor2D,
    CharTensor3D,
    ComplexFloatingTensor,
    ComplexFloatingTensor0D,
    ComplexFloatingTensor1D,
    ComplexFloatingTensor2D,
    ComplexFloatingTensor3D,
    DoubleTensor,
    DoubleTensor0D,
    DoubleTensor1D,
    DoubleTensor2D,
    DoubleTensor3D,
    FloatingTensor,
    FloatingTensor0D,
    FloatingTensor1D,
    FloatingTensor2D,
    FloatingTensor3D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    HalfTensor,
    HalfTensor0D,
    HalfTensor1D,
    HalfTensor2D,
    HalfTensor3D,
    IntTensor,
    IntTensor0D,
    IntTensor1D,
    IntTensor2D,
    IntTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    ShortTensor,
    ShortTensor0D,
    ShortTensor1D,
    ShortTensor2D,
    ShortTensor3D,
    SignedIntegerTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
    SignedIntegerTensor2D,
    SignedIntegerTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)
from .types.variable_fns import as_tensor, empty, full, ones, rand, zeros
from .utils.saving.common import to_builtin
from .utils.saving.csv import load_csv, to_csv
from .utils.saving.json import load_json, to_json
from .utils.saving.pickle import load_pickle, to_pickle

version = __version__
version_info = Version(__version__)

del Version



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch.nn.functional import *

from .activation import log_softmax_multidim, softmax_multidim
from .checksum import checksum, checksum_any
from .crop import crop_dim, crop_dims
from .get import CUDA_IF_AVAILABLE, get_device, get_dtype, get_generator
from .indices import (
    get_inverse_perm,
    get_perm_indices,
    insert_at_indices,
    randperm_diff,
    remove_at_indices,
)
from .mask import (
    generate_square_subsequent_mask,
    lengths_to_non_pad_mask,
    lengths_to_pad_mask,
    masked_equal,
    masked_mean,
    masked_sum,
    non_pad_mask_to_lengths,
    pad_mask_to_lengths,
    tensor_to_lengths,
    tensor_to_non_pad_mask,
    tensor_to_pad_mask,
    tensor_to_tensors_list,
    tensors_list_to_lengths,
)
from .multiclass import (
    index_to_name,
    index_to_onehot,
    name_to_index,
    name_to_onehot,
    onehot_to_index,
    onehot_to_name,
    probs_to_index,
    probs_to_name,
    probs_to_onehot,
)
from .multilabel import (
    indices_to_multihot,
    indices_to_multinames,
    multihot_to_indices,
    multihot_to_multinames,
    multinames_to_indices,
    multinames_to_multihot,
    probs_to_indices,
    probs_to_multihot,
    probs_to_multinames,
)
from .numpy import numpy_to_tensor, tensor_to_numpy, to_numpy
from .others import (
    all_eq,
    all_ne,
    average_power,
    can_be_converted_to_tensor,
    can_be_stacked,
    count_parameters,
    find,
    is_complex,
    is_floating_point,
    is_sorted,
    move_to_rec,
    ndim,
    nelement,
    prod,
    ranks,
    shape,
    to_item,
    view_as_complex,
    view_as_real,
)
from .pad import cat_padded_batch, pad_and_stack_rec, pad_dim, pad_dims
from .powerset import multilabel_to_powerset, powerset_to_multilabel
from .segments import extract_segments, segments_to_list
from .transform import (
    flatten,
    identity,
    pad_and_crop_dim,
    repeat_interleave_nd,
    resample_nearest_freqs,
    resample_nearest_rates,
    resample_nearest_steps,
    shuffled,
    to_tensor,
    transform_drop,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict

from .typing.classes import DataclassInstance
from .typing.guards import is_dataclass_instance  # noqa: F401


def get_defaults_values(dataclass: DataclassInstance) -> Dict[str, Any]:
    defaults = {
        f.name: f.default_factory() if callable(f.default_factory) else f.default
        for f in dataclass.__dataclass_fields__.values()
    }
    return defaults



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional

from torch import Tensor, nn

from torchoutil.nn.functional.powerset import (
    build_powerset_mapping,
    multilabel_to_powerset,
    powerset_to_multilabel,
)
from torchoutil.types import Tensor2D, Tensor3D


class MultilabelToPowerset(nn.Module):
    """
    Module version of :func:`~torchoutil.nn.functional.powerset.multilabel_to_powerset`.
    """

    def __init__(self, num_classes: int, max_set_size: int) -> None:
        mapping = build_powerset_mapping(num_classes, max_set_size)

        super().__init__()
        self._max_set_size = max_set_size

        self.register_buffer("_mapping", mapping, persistent=False)
        self._mapping: Tensor2D

    def forward(self, multilabel: Tensor) -> Tensor:
        return multilabel_to_powerset(multilabel, mapping=self._mapping)

    @property
    def max_set_size(self) -> int:
        return self._max_set_size

    @property
    def num_powerset_classes(self) -> int:
        return self._mapping.shape[0]

    @property
    def num_classes(self) -> int:
        return self._mapping.shape[1]


class PowersetToMultilabel(nn.Module):
    """
    Module version of :func:`~torchoutil.nn.functional.powerset.powerset_to_multilabel`.
    """

    def __init__(self, num_classes: int, max_set_size: int, soft: bool = False) -> None:
        mapping = build_powerset_mapping(num_classes, max_set_size)

        super().__init__()
        self._max_set_size = max_set_size
        self._soft = soft

        self.register_buffer("_mapping", mapping, persistent=False)
        self._mapping: Tensor2D

    def forward(self, powerset: Tensor, soft: Optional[bool] = None) -> Tensor3D:
        if soft is None:
            soft = self._soft
        return powerset_to_multilabel(powerset, soft=soft, mapping=self._mapping)

    @property
    def max_set_size(self) -> int:
        return self._max_set_size

    @property
    def num_powerset_classes(self) -> int:
        return self._mapping.shape[0]

    @property
    def num_classes(self) -> int:
        return self._mapping.shape[1]

    @property
    def soft(self) -> bool:
        return self._soft



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from enum import auto
from typing import Any, Dict, Final

import torch
from torch.types import _bool, _int

from torchoutil.pyoutil.enum import StrEnum

TORCH_DTYPES: Final[Dict[str, torch.dtype]] = {
    # Base types
    "float16": torch.float16,
    "float32": torch.float32,
    "float64": torch.float64,
    "int16": torch.int16,
    "int32": torch.int32,
    "int64": torch.int64,
    "complex32": torch.complex32,
    "complex64": torch.complex64,
    "complex128": torch.complex128,
    # Aliases
    "half": torch.half,
    "float": torch.float,
    "double": torch.double,
    "short": torch.short,
    "int": torch.int,
    "long": torch.long,
    "chalf": torch.chalf if hasattr(torch, "chalf") else torch.complex32,
    "cfloat": torch.cfloat,
    "cdouble": torch.cdouble,
    # Others
    "bfloat16": torch.bfloat16,
    "bool": torch.bool,
    "int8": torch.int8,
    "qint8": torch.qint8,
    "qint32": torch.qint32,
    "quint4x2": torch.quint4x2,
    "quint8": torch.quint8,
    "uint8": torch.uint8,
}

# Optional
if hasattr(torch, "quint2x4"):
    TORCH_DTYPES["quint2x4"] = torch.quint2x4
if hasattr(torch, "uint16"):
    TORCH_DTYPES["uint16"] = torch.uint16
if hasattr(torch, "uint32"):
    TORCH_DTYPES["uint32"] = torch.uint32
if hasattr(torch, "uint64"):
    TORCH_DTYPES["uint64"] = torch.uint64


_NAME_TO_DTYPE: Final[Dict[str, torch.dtype]] = TORCH_DTYPES
_DTYPE_TO_NAME: Final[Dict[torch.dtype, str]] = {
    dt: name for name, dt in _NAME_TO_DTYPE.items()
}


class DTypeEnum(StrEnum):
    """Enum of torch dtypes."""

    # Base types
    float16 = auto()
    float32 = auto()
    float64 = auto()
    int16 = auto()
    int32 = auto()
    int64 = auto()
    complex32 = auto()
    complex64 = auto()
    complex128 = auto()

    # Aliases
    half = float16
    float = float32
    double = float64
    short = int16
    int = int32
    long = int64
    chalf = complex32
    cfloat = complex64
    cdouble = complex128

    # Others
    bfloat16 = auto()
    bool = auto()
    int8 = auto()  # char
    qint8 = auto()
    qint32 = auto()
    quint4x2 = auto()
    quint8 = auto()
    uint8 = auto()  # byte

    # Optional
    quint2x4 = auto()
    uint16 = auto()
    uint32 = auto()
    uint64 = auto()

    @classmethod
    def default(cls) -> "DTypeEnum":
        return cls.from_dtype(torch.get_default_dtype())

    @classmethod
    def from_dtype(cls, dtype: torch.dtype) -> "DTypeEnum":
        if dtype not in _DTYPE_TO_NAME:
            msg = f"Invalid argument {dtype=}. (expected one of {tuple(_DTYPE_TO_NAME.keys())})"
            raise ValueError(msg)
        return DTypeEnum.from_str(_DTYPE_TO_NAME[dtype])

    @property
    def dtype(self) -> torch.dtype:
        return TORCH_DTYPES[self.name]

    @property
    def is_floating_point(self) -> _bool:
        return self.dtype.is_floating_point

    @property
    def is_complex(self) -> _bool:
        return self.dtype.is_complex

    @property
    def is_signed(self) -> _bool:
        return self.dtype.is_signed

    @property
    def itemsize(self) -> _int:
        return self.dtype.itemsize

    def to_real(self) -> "DTypeEnum":
        return DTypeEnum.from_dtype(self.dtype.to_real())

    def to_complex(self) -> "DTypeEnum":
        return DTypeEnum.from_dtype(self.dtype.to_complex())

    def __eq__(self, other: Any) -> _bool:
        if isinstance(other, DTypeEnum):
            return self.dtype == other.dtype
        elif isinstance(other, torch.dtype):
            return self.dtype == other
        elif isinstance(other, str):
            return self.dtype == str_to_torch_dtype(other)
        else:
            return False

    def __hash__(self) -> _int:
        return hash(self.dtype)


def torch_dtype_to_str(dtype: torch.dtype) -> str:
    return _removeprefix(str(dtype), "torch.")


def str_to_torch_dtype(dtype: str) -> torch.dtype:
    return _NAME_TO_DTYPE[_removeprefix(dtype, "torch.")]


def torch_dtype_to_enum_dtype(dtype: torch.dtype) -> DTypeEnum:
    return DTypeEnum.from_dtype(dtype)


def str_to_enum_dtype(dtype: str) -> DTypeEnum:
    return DTypeEnum.from_str(dtype)


def enum_dtype_to_str(dtype: DTypeEnum) -> str:
    return str(dtype)


def enum_dtype_to_torch_dtype(dtype: DTypeEnum) -> torch.dtype:
    return dtype.dtype


def _removeprefix(x: str, prefix: str) -> str:
    # str.removeprefix does not exists in 3.8, so we use this function instead
    if x.startswith(prefix):
        return x[len(prefix) :]
    else:
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, List, Literal, Optional, Tuple, TypedDict, TypeVar

T = TypeVar("T", covariant=True)

ContentMode = Literal["item", "batch", "column"]
ItemType = Literal["dict", "tuple", "raw"]
ExistsMode = Literal["overwrite", "skip", "error"]


EXISTS_MODES = ("overwrite", "skip", "error")
CONTENT_DNAME = "data"
ATTRS_FNAME = "attributes.json"


class PackedDatasetAttributes(TypedDict):
    batch_size: int
    content_dname: str
    content_mode: ContentMode
    creation_date: str
    files: List[str]
    info: Dict[str, Any]
    length: int
    num_files: int
    source_attrs: Dict[str, Any]
    source_dataset: str
    subdir_size: Optional[int]
    item_type: ItemType


def _tuple_to_dict(x: Tuple[T, ...]) -> Dict[str, T]:
    return dict(zip(map(str, range(len(x))), x))


def _dict_to_tuple(x: Dict[str, T]) -> Tuple[T, ...]:
    return tuple(x.values())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import random
import unittest
from unittest import TestCase

from torchoutil.pyoutil.collections import (
    all_eq,
    all_ne,
    dict_list_to_list_dict,
    flat_dict_of_dict,
    flat_list_of_list,
    flatten,
    intersect_lists,
    is_sorted,
    list_dict_to_dict_list,
    unflat_dict_of_dict,
    unflat_list_of_list,
)
from torchoutil.pyoutil.re import get_key_fn
from torchoutil.pyoutil.typing import is_list_list_str, is_list_str


class TestAllEqAllNe(TestCase):
    def test_all_eq(self) -> None:
        x = [1, 2, 3, 4]
        assert not all_eq(x)

        x = [1] * 10
        assert all_eq(x)
        assert all_eq([random.randint(0, 2**20)])

    def test_all_ne(self) -> None:
        lst = [random.randint(0, 10) for _ in range(100)]
        assert all_ne(set(lst))
        assert all_ne([random.randint(0, 2**20)])


class TestDictListToListDict(TestCase):
    def test_example_1(self) -> None:
        x = {"a": [1, 2], "b": [3, 4]}
        expected = [{"a": 1, "b": 3}, {"a": 2, "b": 4}]
        result = dict_list_to_list_dict(x)

        assert result == expected

    def test_example_2(self) -> None:
        x = {"a": [1, 2, 3], "b": [4], "c": [5, 6]}
        expected = [
            {"a": 1, "b": 4, "c": 5},
            {"a": 2, "b": -1, "c": 6},
            {"a": 3, "b": -1, "c": -1},
        ]
        result = dict_list_to_list_dict(x, key_mode="union", default_val=-1)

        assert result == expected


class TestFlatDict(TestCase):
    def test_example_1(self) -> None:
        x = {
            "a": 1,
            "b": {
                "a": 2,
                "b": 10,
            },
        }
        expected = {"a": 1, "b.a": 2, "b.b": 10}

        output = flat_dict_of_dict(x)
        assert output == expected

    def test_example_2(self) -> None:
        x = {"a": ["hello", "world"], "b": 3}
        expected = {"a.0": "hello", "a.1": "world", "b": 3}

        output = flat_dict_of_dict(x, flat_iterables=True)
        assert output == expected

    def test_example_3(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        expected = {"a.b": 2}

        output = flat_dict_of_dict(x)
        assert output == expected

    def test_example_4(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        self.assertRaises(ValueError, flat_dict_of_dict, x, overwrite=False)

    def test_unflat_dict_example_1(self) -> None:
        x = {
            "a.a": 1,
            "b.a": 2,
            "b.b": 3,
            "c": 4,
        }
        y = {"a": {"a": 1}, "b": {"a": 2, "b": 3}, "c": 4}

        y_hat = unflat_dict_of_dict(x)
        assert y_hat == y


class TestListDictToDictList(TestCase):
    def test_example_1(self) -> None:
        lst = [{"a": 1, "b": 2}, {"a": 4, "b": 3, "c": 5}]
        output = list_dict_to_dict_list(
            lst,
            key_mode="union",
            default_val=0,
        )
        expected = {"a": [1, 4], "b": [2, 3], "c": [0, 5]}
        assert output == expected

    def test_example_2(self) -> None:
        lst = [{"a": 1, "b": 2}, {"a": 4, "b": 3, "c": 5}]

        with self.assertRaises(ValueError):
            list_dict_to_dict_list(
                lst,
                default_val=None,
                key_mode="same",
            )

    def test_example_3(self) -> None:
        lst = [{"a": 1, "b": 2, "c": 3}, {"a": 11, "b": 22, "c": 33}]
        output = list_dict_to_dict_list(
            lst,
            default_val=None,
            key_mode="same",
        )
        expected = {"a": [1, 11], "b": [2, 22], "c": [3, 33]}
        assert output == expected


class TestIntersectLists(TestCase):
    def test_example_1(self) -> None:
        input_ = [["a", "b", "b", "c"], ["c", "d", "b", "a"], ["b", "a", "a", "e"]]
        expected = ["a", "b"]
        output = intersect_lists(input_)
        assert output == expected


class TestFlatList(TestCase):
    def test_example_1(self) -> None:
        lst = [
            list(map(str, range(random.randint(0, 100))))
            for _ in range(random.randint(0, 10))
        ]
        for sublst in lst:
            random.shuffle(sublst)
        random.shuffle(lst)

        assert is_list_list_str(lst)

        flatten, sizes = flat_list_of_list(lst)
        assert is_list_str(flatten)
        assert len(lst) == len(sizes)
        assert len(flatten) == sum(sizes)

        unflat = unflat_list_of_list(flatten, sizes)
        assert is_list_list_str(unflat)
        assert len(lst) == len(unflat)
        assert lst == unflat


class TestGetKeyFn(TestCase):
    def test_example_1(self) -> None:
        lst = ["a", "abc", "aa", "abcd"]
        patterns = ["^ab.*"]  # sort list with elements starting with 'ab' first
        result = list(sorted(lst, key=get_key_fn(patterns)))
        expected = ["abc", "abcd", "a", "aa"]
        assert result == expected


class TestFlatten(TestCase):
    def test_example_1(self) -> None:
        xlst = [[[3.0, 0, 1], ["a", None, 2], range(3)]]
        expected = [3.0, 0, 1, "a", None, 2, 0, 1, 2]
        result = flatten(xlst)
        assert result == expected

    def test_example_2(self) -> None:
        xlst = [[range(0, 3), range(3, 6)], [range(6, 9), range(9, 12)]]

        expected = list(range(0, 12))
        result = flatten(xlst)
        assert result == expected

        expected = [list(range(0, 6)), list(range(6, 12))]
        result = flatten(xlst, 1, 2)
        assert result == expected

        expected = [range(0, 3), range(3, 6), range(6, 9), range(9, 12)]
        result = flatten(xlst, 0, 1)
        assert result == expected


class TestIsSorted(TestCase):
    def test_example_1(self) -> None:
        x_sorted = list(range(10))
        x = copy.copy(x_sorted)
        x[0] = x_sorted[-1]
        x[-1] = x_sorted[0]

        assert is_sorted(x_sorted)
        assert not is_sorted(x)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Generator, Tensor, nn

from torchoutil.core.get import DeviceLike, DTypeLike
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.transform import (
    PadCropAlign,
    PadMode,
    PadValue,
    T_BuiltinScalar,
    flatten,
    identity,
    pad_and_crop_dim,
    repeat_interleave_nd,
    resample_nearest_freqs,
    resample_nearest_rates,
    resample_nearest_steps,
    shuffled,
    to_tensor,
    transform_drop,
)
from torchoutil.pyoutil.collections import dump_dict

T = TypeVar("T")


class RepeatInterleaveNd(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.repeat_interleave_nd`.
    """

    def __init__(self, repeats: int, dim: int) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return repeat_interleave_nd(x, self.repeats, self.dim)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats, dim=self.dim))


class ResampleNearestRates(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_rates`.
    """

    def __init__(
        self,
        rates: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        super().__init__()
        self.rates = rates
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_rates(
            x,
            rates=self.rates,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(rates=self.rates, dims=self.dims))


class ResampleNearestFreqs(nn.Module):
    def __init__(
        self,
        orig_freq: int,
        new_freq: int,
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_freqs`.
        """
        super().__init__()
        self.orig_freq = orig_freq
        self.new_freq = new_freq
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_freqs(
            x,
            orig_freq=self.orig_freq,
            new_freq=self.new_freq,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(orig_freq=self.orig_freq, new_freq=self.new_freq, dims=self.dims)
        )


class ResampleNearestSteps(nn.Module):
    def __init__(
        self,
        steps: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_steps`.
        """
        super().__init__()
        self.steps = steps
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_steps(
            x,
            steps=self.steps,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(steps=self.steps, dims=self.dims))


class TransformDrop(Generic[T], nn.Module):
    def __init__(
        self,
        transform: Callable[[T], T],
        p: float,
        generator: Union[int, Generator, None] = None,
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.transform_drop`.
        """
        super().__init__()
        self.transform = transform
        self.p = p
        self.generator = generator

    def forward(self, x: T) -> T:
        return transform_drop(
            transform=self.transform,
            x=x,
            p=self.p,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(p=self.p))


class PadAndCropDim(nn.Module):
    def __init__(
        self,
        target_length: int,
        align: PadCropAlign = "left",
        pad_value: PadValue = 0.0,
        dim: int = -1,
        mode: PadMode = "constant",
        generator: Union[int, Generator, None] = None,
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.pad_and_crop_dim`.
        """
        super().__init__()
        self.target_length = target_length
        self.align: PadCropAlign = align
        self.pad_value = pad_value
        self.dim = dim
        self.mode: PadMode = mode
        self.generator = generator

    def forward(self, x: Tensor) -> Tensor:
        return pad_and_crop_dim(
            x,
            self.target_length,
            align=self.align,
            pad_value=self.pad_value,
            dim=self.dim,
            mode=self.mode,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_length=self.target_length,
                align=self.align,
                pad_value=self.pad_value,
                dim=self.dim,
                mode=self.mode,
            )
        )


class Shuffled(nn.Module):
    def __init__(
        self,
        dims: Union[int, Iterable[int]],
        generator: Union[int, Generator, None],
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.shuffled`.
        """
        super().__init__()
        self.dims = dims
        self.generator = generator

    def forward(self, x: Tensor) -> Tensor:
        return shuffled(x, dims=self.dims, generator=self.generator)

    def extra_repr(self) -> str:
        return dump_dict(dict(dims=self.dims))


class Flatten(nn.Module):
    def __init__(self, start_dim: int = 0, end_dim: Optional[int] = None) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.flatten`.
        """
        super().__init__()
        self.start_dim = start_dim
        self.end_dim = end_dim

    @overload
    def forward(self, x: Tensor) -> Tensor:
        ...

    @overload
    def forward(self, x: Union[np.ndarray, np.generic]) -> np.ndarray:
        ...

    @overload
    def forward(self, x: T_BuiltinScalar) -> List[T_BuiltinScalar]:
        ...

    @overload
    def forward(self, x: Iterable[T_BuiltinScalar]) -> List[T_BuiltinScalar]:
        ...

    @overload
    def forward(self, x: Any) -> List[Any]:
        ...

    def forward(self, x: Any) -> Any:
        return flatten(
            x,
            start_dim=self.start_dim,
            end_dim=self.end_dim,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                start_dim=self.start_dim,
                end_dim=self.end_dim,
            )
        )


class Identity(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        """Identity class placeholder.

        Unlike torch.nn.Identity which only supports Tensor typing, its type output is the same than its input type.
        """
        super().__init__()

    def forward(self, x: T) -> T:
        return identity(x)


class ToTensor(nn.Module):
    def __init__(
        self,
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = None,
    ) -> None:
        """
        For more information, see :func:`~torchoutil.nn.functional.transform.to_tensor`.
        """
        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: Any) -> Tensor:
        return to_tensor(x, dtype=self.dtype, device=self.device)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dtype=self.dtype,
                device=self.device,
            ),
            ignore_lst=(None,),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch

from torchoutil.pyoutil.os import get_num_cpus_available


def get_auto_num_cpus() -> int:
    """Returns the number of CPUs available for the process on Linux-based platforms.
    Useful for setting num_workers argument in DataLoaders.

    On Windows and MAC OS, this will just return the number of CPUs on this machine.
    If the number of CPUs cannot be detected, returns 0.

    Alias of `pyoutil.os.get_num_cpus`.
    """
    return get_num_cpus_available()


def get_auto_num_gpus() -> int:
    """Returns the number of GPUs available to the current process."""
    return torch.cuda.device_count()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime

ISO8601_DAY_FORMAT = r"%Y-%m-%d"
ISO8601_HOUR_FORMAT = r"%Y-%m-%dT%H:%M:%S"


def now_iso(fmt: str = ISO8601_HOUR_FORMAT) -> str:
    return datetime.datetime.now().strftime(fmt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .definitions import ACCEPTED_NUMPY_DTYPES, np, numpy
from .functional import (
    is_numpy_bool_array,
    is_numpy_number_like,
    is_numpy_scalar_like,
    logical_and_lst,
    logical_or_lst,
    numpy_all_eq,
    numpy_all_ne,
    numpy_complex_dtype_to_float_dtype,
    numpy_is_complex,
    numpy_is_complex_dtype,
    numpy_is_floating_point,
    numpy_item,
    numpy_to_tensor,
    numpy_topk,
    numpy_view_as_complex,
    numpy_view_as_real,
    tensor_to_numpy,
    to_numpy,
)
from .modules import NumpyToTensor, TensorToNumpy, ToNumpy
from .scan_info import (
    InvalidTorchDType,
    ShapeDTypeInfo,
    get_default_numpy_dtype,
    merge_numpy_dtypes,
    merge_torch_dtypes,
    numpy_dtype_to_fill_value,
    numpy_dtype_to_torch_dtype,
    scan_numpy_dtype,
    scan_shape_dtypes,
    scan_torch_dtype,
    torch_dtype_to_numpy_dtype,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import unittest
from typing import Iterable, Literal, Mapping
from unittest import TestCase

from torchoutil.pyoutil import inspect
from torchoutil.pyoutil.inspect import get_fullname


class DummyClass:
    def f(self):
        return 0


class TestPyoutilInspect(TestCase):
    def test_example_1(self) -> None:
        x = [0, 1, 2]
        assert get_fullname(x) == "builtins.list(...)"

        x = 1.0
        assert get_fullname(x) == "builtins.float(...)"

        x = DummyClass()
        assert get_fullname(x) == f"{self.__module__}.DummyClass(...)"
        assert get_fullname(x.f) == f"{self.__module__}.DummyClass.f"

        assert get_fullname(DummyClass) == f"{self.__module__}.DummyClass"
        assert get_fullname(DummyClass.f) == f"{self.__module__}.DummyClass.f"

    def test_example_2(self) -> None:
        assert get_fullname(TestCase) == "unittest.case.TestCase"
        assert get_fullname(inspect) == "torchoutil.pyoutil.inspect"
        assert get_fullname(get_fullname) == "torchoutil.pyoutil.inspect.get_fullname"

        if sys.version_info.minor >= 11:
            assert get_fullname(Mapping) == "typing.Mapping"
            assert get_fullname(Iterable[str]) == "typing.Iterable[builtins.str]"
            assert (
                get_fullname(Iterable[Literal[1]])
                == "typing.Iterable[typing.Literal[builtins.int(...)]]"
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, List, Literal, Union

import torch
from torch import Generator, Tensor

from torchoutil.core.get import get_generator

CROP_ALIGNS = ("left", "right", "center", "random")
CropAlign = Literal["left", "right", "center", "random"]


def crop_dim(
    x: Tensor,
    target_length: int,
    *,
    dim: int = -1,
    align: CropAlign = "left",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function to crop a single dimension."""
    return crop_dims(
        x,
        [target_length],
        dims=[dim],
        aligns=[align],
        generator=generator,
    )


def crop_dims(
    x: Tensor,
    target_lengths: Iterable[int],
    *,
    dims: Union[Iterable[int], Literal["auto"]] = "auto",
    aligns: Union[CropAlign, Iterable[CropAlign]] = "left",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Generic function to crop multiple dimensions."""

    target_lengths = list(target_lengths)

    aligns_lst: List[CropAlign]
    if isinstance(aligns, str):
        aligns_lst = [aligns] * len(target_lengths)
    else:
        aligns_lst = list(aligns)
    del aligns

    if dims == "auto":
        dims = list(range(-len(target_lengths), 0))
    else:
        dims = list(dims)

    generator = get_generator(generator)

    if len(target_lengths) != len(dims):
        msg = f"Invalid number of targets lengths ({len(target_lengths)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    if len(aligns_lst) != len(dims):
        msg = f"Invalid number of aligns ({len(aligns_lst)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    slices = [slice(None)] * len(x.shape)

    for target_length, dim, align in zip(target_lengths, dims, aligns_lst):
        if x.shape[dim] <= target_length:
            continue

        if align == "left":
            start = 0
            end = target_length
        elif align == "right":
            start = x.shape[dim] - target_length
            end = None
        elif align == "center":
            diff = x.shape[dim] - target_length
            start = diff // 2 + diff % 2
            end = start + target_length
        elif align == "random":
            diff = x.shape[dim] - target_length
            start = torch.randint(low=0, high=diff, size=(), generator=generator).item()
            end = start + target_length
        else:
            msg = f"Invalid argument {align=}. (expected one of {CROP_ALIGNS})"
            raise ValueError(msg)

        slices[dim] = slice(start, end)

    x = x[slices]
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchoutil.core.get import (  # noqa: F401
    CUDA_IF_AVAILABLE,
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    get_device,
    get_dtype,
    get_generator,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import merge_numpy_dtypes, np


class TestDType(TestCase):
    def test_merge_numpy_dtypes(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        empty = None
        args_lst = [
            [np.int16, np.int32, np.int64],
            [np.complex64, np.float16, np.float64],
            [],
            [np.int64, np.float16],
            [np.dtype("<U2"), np.dtype("<U10"), empty],
        ]
        expected_lst = [
            np.int64,
            np.complex128,
            empty,
            np.float64,
            np.dtype("<U10"),
        ]

        for args, expected in zip(args_lst, expected_lst):
            result = merge_numpy_dtypes(args, empty=empty)
            assert result == expected, f"{result=}; {expected=}"

    def test_result_dtype_compat(self) -> None:
        args_lst = [
            [np.int16, np.int32, np.int64],
            [np.complex64, np.float16, np.float64],
            [np.int64, np.float16],
            [np.dtype("<U2"), np.dtype("<U10")],
            [np.int32, np.object_],
        ]

        for args in args_lst:
            result = merge_numpy_dtypes(args)
            expected = np.result_type(*args)
            assert result == expected


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from collections import Counter
from pathlib import Path
from unittest import TestCase

import torch

from torchoutil.utils.saving import to_builtin


class TestSaving(TestCase):
    def test_examples(self) -> None:
        x = [
            [
                torch.arange(3)[None],
                "a",
                Path("./path"),
                Counter(["a", "b", "a", "c", "a"]),
                (),
            ],
        ]
        expected = [[[list(range(3))], "a", "path", {"a": 3, "b": 1, "c": 1}, []]]
        result = to_builtin(x)
        assert result == expected, f"{result=}; {expected=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

import torchoutil as to
from torchoutil.core.semver import Version


class TestVersion(TestCase):
    def test_versions(self) -> None:
        v1 = Version("1.0.0")
        assert v1.to_dict() == {"major": 1, "minor": 0, "patch": 0}

        v2 = Version("1.1.0")
        assert v2.to_dict() == {"major": 1, "minor": 1, "patch": 0}

        assert v1 != v2
        assert v1 < v2
        assert not (v1 > v2)

        v3 = Version("1.2.0")
        assert v3.to_dict() == {"major": 1, "minor": 2, "patch": 0}

        v4 = Version(minor=3, patch=10, major=2)
        assert v4.to_str() == "2.3.10"

        v5 = Version((2, 3, 10))
        assert v5.to_str() == "2.3.10"
        assert v4 == v5

        v6 = Version((3, 4, 5))
        v7 = Version(3, 4, 5)
        assert v6 == v7
        assert v6.to_str() == "3.4.5"

        v10 = Version("1.2.0+test")
        assert v10.to_dict() == {
            "major": 1,
            "minor": 2,
            "patch": 0,
            "buildmetadata": "test",
        }

        # Check if versions can be parsed
        Version(to.__version__)
        Version(torch.__version__)

    def test_parse_invalid(self) -> None:
        with self.assertRaises(ValueError):
            Version("")

        with self.assertRaises(ValueError):
            Version(".")

        with self.assertRaises(ValueError):
            Version("1.")

        with self.assertRaises(ValueError):
            Version("1.2")

        with self.assertRaises(ValueError):
            Version("1..0")

        with self.assertRaises(ValueError):
            Version("1.None")

        with self.assertRaises(ValueError):
            Version("1.2.X")

    def test_semver(self) -> None:
        assert Version("1.0.0") < Version("2.0.0") < Version("2.1.0") < Version("2.1.1")
        assert Version("1.0.0-alpha") < Version("1.0.0")
        assert (
            Version("0.9.0")
            < Version("1.0.0-alpha")
            < Version("1.0.0-alpha.1")
            < Version("1.0.0-alpha.beta")
            < Version("1.0.0-beta")
            < Version("1.0.0-beta.2")
            < Version("1.0.0-beta.11")
            < Version("1.0.0-rc.1")
            < Version("1.0.0")
            < Version("2.0.0")
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import tempfile
import unittest
from pathlib import Path
from unittest import TestCase

import torch

from torchoutil import Tensor, nn
from torchoutil.nn.modules.mixins import _DEFAULT_DEVICE_DETECT_MODE


class Intermediate(torch.nn.Module):
    """Subclass of torch Module instead of torchoutil EModule here."""

    def __init__(self, in_features: int, out_features: int) -> None:
        super().__init__()
        self.linear = nn.Linear(in_features, out_features)
        self.intermediate_attr = "abcd"

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(x)


class MyModule(nn.EModule[Tensor, Tensor]):
    def __init__(self, in_features: int, out_features: int, p: float) -> None:
        projection = Intermediate(in_features, out_features)
        dropout = nn.Dropout(p=p)

        super().__init__()
        self.projection = projection
        self.dropout = dropout
        self.p = p
        self.sizes = (in_features, out_features)

    def forward(self, x: Tensor) -> Tensor:
        return self.dropout(self.projection(x))


class TestInheritEModule(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        cls.tmpdir = tmpdir

    def test_multiple(self) -> None:
        bsize = 8
        in_features = 2**5
        out_features = 2**4

        module1 = MyModule(in_features, out_features, p=0.5)
        x = torch.rand(bsize, in_features)
        result = module1(x)

        assert tuple(result.shape) == (bsize, out_features)
        assert module1.get_device() == torch.device("cpu")

        expected_config = {
            "projection.linear.in_features": 32,
            "projection.linear.out_features": 16,
            "projection.intermediate_attr": "abcd",
            "dropout.p": 0.5,
            "dropout.inplace": False,
            "p": 0.5,
            "sizes": (32, 16),
        }
        assert module1.config == expected_config
        assert module1.count_parameters() == in_features * out_features + out_features

        path = self.tmpdir.joinpath("state_dict.pt")
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as file:
            torch.save(module1.state_dict(), file)
            state_dict = torch.load(path)

        module2 = MyModule(in_features, out_features, p=0.25)
        module2.load_state_dict(state_dict)

        assert module1.checksum() == module2.checksum()

    def test_example_device_detect_mode_default(self) -> None:
        module = MyModule(10, 16, p=0.0)
        assert module.device_detect_mode == _DEFAULT_DEVICE_DETECT_MODE


if __name__ == "__main__":
    unittest.main()
